#!/usr/bin/env python
# -*- coding: utf-8 -*-

# netgraph.py --- Python drawing utilities for publication quality plots of networks.

# Copyright (C) 2016 Paul Brodersen <paulbrodersen+netgraph@gmail.com>

# Author: Paul Brodersen <paulbrodersen+netgraph@gmail.com>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Netgraph
========

Summary
-------
Python drawing utilities for publication quality plots of networks.

Examples
--------
>>> import matplotlib.pyplot as plt
>>> from netgraph import Graph, InteractiveGraph, EditableGraph
>>>
>>> # Several graph formats are supported:
>>>
>>> # 1) edge lists
>>> graph_data = [(0, 1), (1, 2), (2, 0)]
>>>
>>> # 2) edge list with weights
>>> graph_data = [(0, 1, 0.2), (1, 2, -0.4), (2, 0, 0.7)]
>>>
>>> # 3) full rank matrices
>>> import numpy
>>> graph_data = np.random.rand(10, 10)
>>>
>>> # 4) networkx Graph and DiGraph objects (MultiGraph objects are not supported, yet)
>>> import networkx
>>> graph_data = networkx.karate_club_graph()
>>>
>>> # 5) igraph.Graph objects
>>> import igraph
>>> graph_data = igraph.Graph.Famous('Zachary')
>>>
>>> # 6) graph_tool.Graph objects
>>> import graph_tool.collection
>>> graph_data = graph_tool.collection.data["karate"]
>>>
>>> # Create a non-interactive plot:
>>> Graph(graph_data)
>>> plt.show()
>>>
>>> # Create an interactive plot, in which the nodes can be re-positioned with the mouse.
>>> # NOTE: you must retain a reference to the plot instance!
>>> # Otherwise, the plot instance will be garbage collected after the initial draw
>>> # and you won't be able to move the plot elements around.
>>> # For related reasons, if you are using PyCharm, you have to execute the code in
>>> # a console (Alt+Shift+E).
>>> plot_instance = InteractiveGraph(graph_data)
>>> plt.show()
>>>
>>> # Create an editable plot, which is an interactive plot with the additions
>>> # that nodes and edges can be inserted or deleted, and labels and annotations
>>> # can be created, edited, or deleted as well.
>>> plot_instance = EditableGraph(graph_data)
>>> plt.show()
>>>
>>> # Netgraph uses matplotlib for creating the visualisation.
>>> # Node and edge artistis are derived from `matplotlib.patches.PathPatch`.
>>> # Node and edge labels are `matplotlib.text.Text` instances.
>>> # Standard matplotlib syntax applies.
>>> fig, ax = plt.subplots(figsize=(5,4))
>>> plot_instance = Graph([(0, 1)], node_labels=True, edge_labels=True, ax=ax)
>>> plot_instance.node_artists[0].set_alpha(0.2)
>>> plot_instance.edge_artists[(0, 1)].set_facecolor('red')
>>> plot_instance.edge_label_artists[(0, 1)].set_style('italic')
>>> plot_instance.node_label_artists[1].set_size(10)
>>> ax.set_title("This is my fancy title.")
>>> ax.set_facecolor('honeydew') # change background color
>>> fig.canvas.draw() # force redraw to display changes
>>> fig.savefig('test.pdf', dpi=300)
>>> plt.show()
>>>
>>> # Read the documentation for a full list of available arguments:
>>> help(Graph)
>>> help(InteractiveGraph)
>>> help(EditableGraph)
"""

__version__ = "4.12.10"
__author__ = "Paul Brodersen"
__email__ = "paulbrodersen+netgraph@gmail.com"

from ._graph_classes import (
    BASE_SCALE,
    DEFAULT_COLOR,
    BaseGraph,
    Graph,
)

from ._interactive_graph_classes import (
    InteractiveGraph,
    MutableGraph,
    EditableGraph,
)

from ._multigraph_classes import (
    BaseMultiGraph,
    MultiGraph,
)

from ._interactive_multigraph_classes import (
    InteractiveMultiGraph,
    MutableMultiGraph,
    EditableMultiGraph,
)

from ._arcdiagram_classes import (
    BaseArcDiagram,
    ArcDiagram,
)

from ._interactive_arcdiagram_classes import (
    InteractiveArcDiagram,
    MutableArcDiagram,
    EditableArcDiagram,
)

from ._node_layout import (
    get_random_layout,
    get_fruchterman_reingold_layout,
    get_sugiyama_layout,
    get_radial_tree_layout,
    get_circular_layout,
    get_linear_layout,
    get_bipartite_layout,
    get_multipartite_layout,
    get_shell_layout,
    get_community_layout,
    get_geometric_layout,
)

from ._edge_layout import (
    get_straight_edge_paths,
    get_curved_edge_paths,
    get_arced_edge_paths,
    get_bundled_edge_paths,
)

from ._parser import parse_graph


__all__ = [
    'BASE_SCALE',
    'DEFAULT_COLOR',
    'BaseGraph',
    'Graph',
    'get_random_layout',
    'get_fruchterman_reingold_layout',
    'get_sugiyama_layout',
    'get_radial_tree_layout',
    'get_circular_layout',
    'get_linear_layout',
    'get_bipartite_layout',
    'get_multipartite_layout',
    'get_shell_layout',
    'get_community_layout',
    'get_geometric_layout',
    'get_straight_edge_paths',
    'get_curved_edge_paths',
    'get_arced_edge_paths',
    'get_bundled_edge_paths',
    'InteractiveGraph',
    'MutableGraph',
    'EditableGraph',
    'BaseMultiGraph',
    'MultiGraph',
    'InteractiveMultiGraph',
    'MutableMultiGraph',
    'EditableMultiGraph',
    'BaseArcDiagram',
    'ArcDiagram',
    'InteractiveArcDiagram',
    'MutableArcDiagram',
    'EditableArcDiagram',
    'parse_graph',
]



"""
This module provides various classes for the creation of log during the causal discovery analysis.

Classes:
    CPLevel: support class for handling different levels of log.
    CPrinter: class responsible for the log.
"""

from enum import Enum

class CPLevel(Enum):
    """CPLevel Enumerator."""
    
    NONE = 0
    WARN = 1
    INFO = 2
    DEBUG = 3


class CPrinter():
    """CPrinter class."""
    
    def __init__(self):
        """Class constructor."""
        self.verbosity = None
    
    def set_verbosity(self, verbosity: CPLevel):
        """
        Set verbosity level.

        Args:
            verbosity (CPLevel): verbosity level.
        """
        self.verbosity = verbosity
        
    def warning(self, msg: str):
        """
        Write message iff verbosity >= WARN.

        Args:
            msg (str): massage to write.
        """
        if self.verbosity.value >= CPLevel.WARN.value: print(msg)

    def info(self, msg: str):
        """
        Write message iff verbosity >= INFO.

        Args:
            msg (str): massage to write.
        """
        if self.verbosity.value >= CPLevel.INFO.value: print(msg)

    def debug(self, msg: str):
        """
        Write message iff verbosity >= DEBUG.

        Args:
            msg (str): massage to write.
        """
        if self.verbosity.value >= CPLevel.DEBUG.value: print(msg)

CP = CPrinter()


"""
This module provides the CausalDiscoveryMethod class.

Classes:
    CausalDiscoveryMethod: abstract class used by all the causal discovery algorithms.
"""

from abc import ABC, abstractmethod
import sys
import copy
import pickle
from causalflow.graph.DAG import DAG
from causalflow.CPrinter import CPLevel, CP
from causalflow.basics.constants import *
from causalflow.basics.logger import Logger
import causalflow.basics.utils as utils
from causalflow.preprocessing.data import Data


class CausalDiscoveryMethod(ABC):
    """
    CausalDiscoveryMethod class.

    CausalDiscoveryMethod is an abstract causal discovery method for 
    large-scale time series datasets.
    """

    def __init__(self, 
                 data: Data, 
                 min_lag, max_lag, 
                 verbosity: CPLevel, 
                 alpha = 0.05, 
                 resfolder = None,
                 neglect_only_autodep = False,
                 clean_cls = True):
        """
        Class contructor.

        Args:
            data (Data): data to analyse.
            min_lag (int): minimum time lag.
            max_lag (int): maximum time lag.
            verbosity (CPLevel): verbosity level.
            alpha (float, optional): significance level. Defaults to 0.05.
            resfolder (string, optional): result folder to create. Defaults to None.
            neglect_only_autodep (bool, optional): Bit for neglecting variables with only autodependency. Defaults to False.
            clean_cls (bool): Clean console bit. Default to True.

        """
        self.data = data
        self.alpha = alpha
        self.min_lag = min_lag
        self.max_lag = max_lag
        self.CM = DAG(self.data.features, min_lag, max_lag, neglect_only_autodep)
        self.neglect_only_autodep = neglect_only_autodep

        self.resfolder = resfolder
        self.respath, self.dag_path, self.ts_dag_path = None, None, None
        if resfolder is not None:
            logpath, self.respath, self.dag_path, self.ts_dag_path = utils.get_selectorpath(resfolder)  
            self.logger = Logger(logpath, clean_cls)
            sys.stdout = self.logger
        
        CP.set_verbosity(verbosity)


    @abstractmethod
    def run(self) -> DAG:
        """
        Run causal discovery method.
        
        Returns:
            DAG: causal model.
        """
        pass

    
    def load(self, res_path):
        """
        Load previously estimated result .

        Args:
            res_path (str): pickle file path.
        """
        with open(res_path, 'rb') as f:
            r = pickle.load(f)
            self.CM = r['causal_model']
            self.alpha = r['alpha']
            self.dag_path = r['dag_path']
            self.ts_dag_path = r['ts_dag_path']
            
            
    def save(self):
        """Save causal discovery result as pickle file if resfolder is set."""
        if self.respath is not None:
            if self.CM:
                res = dict()
                res['causal_model'] = copy.deepcopy(self.CM)
                res['alpha'] = self.alpha
                res['dag_path'] = self.dag_path
                res['ts_dag_path'] = self.ts_dag_path
                with open(self.respath, 'wb') as resfile:
                    pickle.dump(res, resfile)
            else:
                CP.warning("Causal model impossible to save")


"""
This module provides the Node class.

Classes:
    Node: class for facilitating the handling and the creation of nodes for a DAG.
"""

from causalflow.basics.constants import SCORE

class Node():
    """Node class."""
    
    def __init__(self, name, neglect_autodep):
        """
        Class contructor.

        Args:
            name (str): node name.
            neglect_autodep (bool): flag to decide whether to to skip the node if it is only auto-dependent.
        """
        self.name = name
        self.sources = dict()
        self.children = list()
        self.neglect_autodep = neglect_autodep
        self.intervention_node = False        
        self.associated_context = None        
    
    
    @property
    def is_autodependent(self) -> bool:
        """
        Return True if the node is autodependent.

        Returns:
            bool: Returns True if the node is autodependent. Otherwise False.
        """
        return self.name in self.sourcelist
    
    
    @property
    def is_isolated(self) -> bool:
        """
        Return True if the node is isolated.

        Returns:
            bool: Returns True if the node is isolated. Otherwise False.
        """
        if self.neglect_autodep:
            return (self.is_exogenous and not self.has_child) or self.is_only_autodep or self.is_only_autodep_context
        return (self.is_exogenous or self.has_only_context) and not self.has_child
    
    
    @property
    def is_only_autodep(self) -> bool:
        """
        Return True if the node is ONLY auto-dependent.

        Returns:
            bool: Returns True if the node is ONLY auto-dependent. Otherwise False.
        """
        return len(self.sources) == 1 and self.name in self.sourcelist and len(self.children) == 1 and self.name in self.children
    
    
    @property
    def has_only_context(self) -> bool:
        """
        Return True if the node has ONLY the context variable as parent.

        Returns:
            bool: Returns True if the node has ONLY the context variable as parent. Otherwise False.
        """
        return len(self.sources) == 1 and self.associated_context in self.sourcelist
    
    
    @property
    def is_only_autodep_context(self) -> bool:
        """
        Return True if the node has ONLY the context variable and itself as parent.

        Returns:
            bool: Returns True if the node has ONLY the context variable and itself as parent. Otherwise False.
        """
        return len(self.sources) == 2 and self.name in self.sourcelist and self.associated_context in self.sourcelist and len(self.children) == 1 and self.name in self.children
    
    
    @property
    def is_exogenous(self) -> bool:
        """
        Return True if the node has no parents.

        Returns:
            bool: Returns True if the node has no parents. Otherwise False.
        """
        return len(self.sources) == 0
        
        
    @property
    def has_child(self) -> bool:
        """
        Return True if the node has at least one child.

        Returns:
            bool: Returns True if the node has at least one child. Otherwise False.
        """
        return len(self.children) > 0
    
    
    @property
    def sourcelist(self) -> list:
        """
        Return list of source names.

        Returns:
            list(str): Returns list of source names.
        """
        return [s[0] for s in self.sources]
    
    
    @property
    def autodependency_links(self) -> list:
        """
        Return list of autodependency links.

        Returns:
            list: Returns list of autodependency links.

        """
        autodep_links = list()
        if self.is_autodependent:
            for s in self.sources: 
                if s[0] == self.name: 
                    autodep_links.append(s)
        return autodep_links
    
    
    @property
    def get_max_autodependent(self) -> float:
        """
        Return max score of autodependent link.

        Returns:
            float: Returns max score of autodependent link.
        """
        max_score = 0
        max_s = None
        if self.is_autodependent:
            for s in self.sources: 
                if s[0] == self.name:
                    if self.sources[s][SCORE] > max_score: max_s = s
        return max_s


#!/usr/bin/env python
# coding: utf-8

"""
Edge routing routines.
"""

import itertools
import warnings
import numpy as np

from uuid import uuid4
from functools import wraps
from scipy.interpolate import UnivariateSpline
from matplotlib.patches import ConnectionStyle

from ._utils import (
    _bspline,
    _get_n_points_on_a_circle,
    _get_angle,
    _get_unit_vector,
    _edge_list_to_adjacency_list,
    _edge_list_to_adjacency_matrix,
    _get_connected_components,
    _get_orthogonal_unit_vector,
    _normalize_numeric_argument,
    _resample_spline,
    _get_optimal_offsets,
    _get_gradient_and_intercept,
    _is_above_line,
    _reflect_across_line,
    _print_progress_bar,
    _simplify_multigraph,
    _map_multigraph_edges_to_ids,
    _get_parallel_line,
    _shorten_spline_by,
)

from ._node_layout import get_fruchterman_reingold_layout


# for profiling with kernprof/line_profiler
try:
    profile
except NameError:
    profile = lambda x: x


def _handle_multiple_components(layout_function):
    """If the graph contains multiple components, apply the given layout to each component individually."""
    @wraps(layout_function)
    def wrapped_layout_function(edges, node_positions=None, *args, **kwargs):
        adjacency_list = _edge_list_to_adjacency_list(edges, directed=False)
        components = _get_connected_components(adjacency_list)

        if len(components) > 1:
            return _get_layout_for_multiple_components(edges, node_positions, components, layout_function, *args, **kwargs)
        else:
            return layout_function(edges, node_positions, *args, **kwargs)

    return wrapped_layout_function


def _get_layout_for_multiple_components(edges, node_positions, components, layout_function, *args, **kwargs):
    """Partition network into given components and apply the given layout to each component individually."""
    edge_paths = dict()
    for component in components:
        component_edges = [(source, target) for (source, target) in edges if (source in component) and (target in component)]
        component_node_positions = {node : xy for node, xy in node_positions.items() if node in component}
        component_edge_paths = layout_function(component_edges, component_node_positions, *args, **kwargs)
        edge_paths.update(component_edge_paths)
    return edge_paths


def get_straight_edge_paths(edges, node_positions, selfloop_radius=0.1, selfloop_angle=None):
    """Edge routing using straight lines.

    Computes the edge paths, such that edges are represented by
    straight lines connecting the source and target node.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.
    selfloop_radius : dict or float, default 0.1
        Dictionary mapping each self-loop edge to a radius. If float, all self-loops have the same radius.
    selfloop_angle : dict, float, or None
        The starting angle of the self-loop in radians.
        If None, the angle is adjusted to minimize collisions with other nodes and edges.

    Returns
    -------
    edge_paths : dict
        Dictionary mapping each edge to an array of (x, y) coordinates representing its path.

    """

    edge_paths = dict()

    nonloops = [(source, target) for (source, target) in edges if source != target]
    nonloop_edge_paths = _get_straight_nonloop_edge_paths(nonloops, node_positions)
    edge_paths.update(nonloop_edge_paths)

    selfloops = [(source, target) for (source, target) in edges if source == target]
    if selfloops:
        selfloop_radius = _normalize_numeric_argument(selfloop_radius, selfloops, 'selfloop_radius')
        if selfloop_angle is not None: # can be zero!
            selfloop_angle = _normalize_numeric_argument(selfloop_angle, selfloops, 'angle', allow_none=True)
        else:
            selfloop_angle = _get_optimal_selfloop_angles(
                selfloops, selfloop_radius, node_positions, nonloop_edge_paths)
        selfloop_edge_paths = _get_straight_selfloop_edge_paths(
            selfloops, node_positions, selfloop_radius, selfloop_angle)
        edge_paths.update(selfloop_edge_paths)

    return edge_paths


def _get_straight_nonloop_edge_paths(edges, node_positions):
    edge_paths = dict()
    for (source, target) in edges:
        x1, y1 = node_positions[source]
        x2, y2 = node_positions[target]
        edge_paths[(source, target)] = np.c_[[x1, x2], [y1, y2]]
    return edge_paths


def _get_optimal_selfloop_angles(selfloops, selfloop_radius, node_positions, edge_paths, total_samples_per_edge=100):
    anchors = np.array([node_positions[source] for source, _ in selfloops])
    offsets = np.array([selfloop_radius[edge] for edge in selfloops])
    avoid = np.concatenate([_resample_spline(path, total_samples_per_edge) for path in edge_paths.values()], axis=0)
    selfloop_centers = _get_optimal_offsets(anchors, offsets, avoid)
    selfloop_angles = [_get_angle(*center) for center in selfloop_centers]
    return dict(zip(selfloops, selfloop_angles))


def _get_decent_selfloop_angles(selfloops, node_positions):
    # To minimise overlap with other edges, we want the loop to be
    # on the side of the node away from the centroid of the graph.
    selfloop_angles = []
    if len(node_positions) > 1:
        centroid = np.mean(list(node_positions.values()), axis=0)
        for edge in selfloops:
            delta = node_positions[edge[0]] - centroid
            distance = np.linalg.norm(delta)
            unit_vector = delta / distance
            selfloop_angles.append(_get_angle(*unit_vector))
    else: # single node in graph; self-loop points upwards
        for edge in selfloops:
            unit_vector = np.array([0, 1])
            selfloop_angles.append(_get_angle(*unit_vector))

    return dict(zip(selfloops, selfloop_angles))


def _get_straight_selfloop_edge_paths(edges, node_positions, selfloop_radius, selfloop_angle, total_points=100):
    edge_paths = dict()
    for edge in edges:
        edge_paths[edge] = _get_selfloop_path(
            node_positions[edge[0]], selfloop_radius[edge], selfloop_angle[edge], total_points)
    return edge_paths


def _get_selfloop_path(source_position, radius, angle, total_points):
    unit_vector = _get_unit_vector(np.array([np.cos(angle), np.sin(angle)]))
    center = source_position + radius * unit_vector
    # Note: we add pi to the start angle as the start angle lies opposite
    # to the direction in which the self-loop extends.
    return _get_n_points_on_a_circle(
        center, radius, total_points+1,
        _get_angle(*unit_vector) + np.pi,
    )[1:]


def get_curved_edge_paths(edges, node_positions,
                          selfloop_radius               = 0.1,
                          selfloop_angle                = None,
                          origin                        = np.array([0, 0]),
                          scale                         = np.array([1, 1]),
                          k                             = 0.1,
                          initial_temperature           = 0.01,
                          total_iterations              = 50,
                          node_size                     = 0.,
                          bundle_parallel_edges         = True):
    """Edge routing using curved paths that avoid nodes and other edges.

    Computes the edge paths, such that edges are represented by curved
    lines connecting the source and target node. Edges paths avoid
    nodes and each other. The edge layout is determined using the
    Fruchterman-Reingold algorithm.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.
    selfloop_radius : dict or float, default 0.1
        Dictionary mapping each self-loop edge to a radius. If float, all self-loops have the same radius.
    selfloop_angle : dict, float, or None
        The starting angle of the self-loop in radians.
    origin : numpy.array
        A (float x, float y) tuple corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : numpy.array
        A (float x, float y) tuple representing the width and height of the bounding box specifying the extent of the canvas.
    k : float, default 0.1
        Spring constant, which controls the tautness of edges.
        Small values will result in straight connections, large values in bulging arcs.
    total_iterations : int, default 50
        Number of iterations in the Fruchterman-Reingold algorithm.
    initial_temperature: float, default 1.
        Temperature controls the maximum node displacement on each iteration.
        Temperature is decreased on each iteration to eventually force the algorithm
        into a particular solution. The size of the initial temperature determines how
        quickly that happens. Values should be much smaller than the values of `scale`.
    node_size : float or dict
        Dictionary mapping each node to a float, the node size. Used for node avoidance.
    bundle_parallel_edges: boolean, default True
        If True, parallel edges (including bi-directional edges) have the same path.

    Returns
    -------
    edge_paths : dict
        Dictionary mapping each edge to an array of (x, y) coordinates representing its path.

    """

    edge_paths = dict()

    nonloops = [(source, target) for (source, target) in edges if source != target]
    if bundle_parallel_edges:
        parallel_edges = []
        other_edges = []
        for (source, target) in nonloops:
            if (target, source) in nonloops:
                if (target, source) in parallel_edges:
                    pass
                else:
                    parallel_edges.append((source, target))
            else:
                other_edges.append((source, target))
        nonloop_edge_paths = _get_curved_nonloop_edge_paths(
            parallel_edges + other_edges, node_positions, origin, scale, k, initial_temperature,
            total_iterations, node_size, bundle_parallel_edges)
        for (source, target) in parallel_edges:
            nonloop_edge_paths[(target, source)] = nonloop_edge_paths[(source, target)][::-1]
    else:
        nonloop_edge_paths = _get_curved_nonloop_edge_paths(
            nonloops, node_positions, origin, scale, k, initial_temperature,
            total_iterations, node_size, bundle_parallel_edges)
    edge_paths.update(nonloop_edge_paths)

    selfloops = [(source, target) for (source, target) in edges if source == target]
    if selfloops:
        selfloop_radius = _normalize_numeric_argument(selfloop_radius, selfloops, 'selfloop_radius')
        if selfloop_angle is not None: # can be zero!
            selfloop_angle = _normalize_numeric_argument(selfloop_angle, selfloops, 'angle', allow_none=True)
        else:
            selfloop_angle = _get_optimal_selfloop_angles(
                selfloops, selfloop_radius, node_positions, nonloop_edge_paths)
        selfloop_edge_paths = _get_curved_selfloop_edge_paths(
            selfloops, node_positions, selfloop_radius, selfloop_angle,
            origin, scale, k, initial_temperature,
            total_iterations, node_size, nonloop_edge_paths)
        edge_paths.update(selfloop_edge_paths)

    return edge_paths


def _get_curved_nonloop_edge_paths(edges, node_positions, origin, scale,
                                   k, initial_temperature, total_iterations,
                                   node_size, bundle_parallel_edges):

    edge_to_control_points = _initialize_nonloop_control_points(edges, node_positions, scale)

    control_point_positions = _initialize_nonloop_control_point_positions(
        edge_to_control_points, node_positions, bundle_parallel_edges)

    control_point_positions = _optimize_control_point_positions(
        edge_to_control_points, node_positions, control_point_positions,
        origin, scale, k, initial_temperature, total_iterations, node_size,
        bundle_parallel_edges)

    edge_to_path = _get_path_through_control_points(
        edge_to_control_points, node_positions, control_point_positions)

    edge_to_path = _smooth_edge_paths(edge_to_path)

    return edge_to_path


def _initialize_nonloop_control_points(edges, node_positions, scale):
    """Represent each edge with string of control points."""
    edge_to_control_points = dict()
    for source, target in edges:
        edge_length = np.linalg.norm(node_positions[target] - node_positions[source], axis=-1) / np.linalg.norm(scale)
        total_control_points = min(max(int(edge_length * 10), 1), 5) # ensure that there are at least one point but no more than 5
        edge_to_control_points[(source, target)] = [uuid4() for _ in range(total_control_points)]
    return edge_to_control_points


def _initialize_nonloop_control_point_positions(edge_to_control_points, node_positions, bundle_parallel_edges):
    """Initialise the positions of the control points to positions on a straight line between source and target node."""

    control_point_positions = dict()
    for (source, target), control_points in edge_to_control_points.items():
        delta = node_positions[target] - node_positions[source]
        fraction = np.linspace(0, 1, len(control_points)+2)[1:-1]
        positions = fraction[:, np.newaxis] * delta[np.newaxis, :] + node_positions[source]
        if (not bundle_parallel_edges) and ((target, source) in edge_to_control_points):
            # Offset the path ever so slightly to a side, such that bi-directional edges do not overlap completely.
            # This prevents an intertwining of parallel edges.
            offset = 1e-3 * np.linalg.norm(delta) * np.squeeze(_get_orthogonal_unit_vector(np.atleast_2d(delta)))
            positions -= offset
        control_point_positions.update(zip(control_points, positions))
    return control_point_positions


def _get_curved_selfloop_edge_paths(edges, node_positions, selfloop_radius, selfloop_angle,
                                    origin, scale, k, initial_temperature, total_iterations, node_size,
                                    nonloop_edge_paths):

    edge_to_control_points = _initialize_selfloop_control_points(edges)

    control_point_positions = _initialize_selfloop_control_point_positions(
        edge_to_control_points, node_positions, selfloop_radius, selfloop_angle)

    expanded_node_positions = node_positions.copy()
    for positions in nonloop_edge_paths.values():
        expanded_node_positions.update(zip([uuid4() for _ in range(len(positions)-2)], positions[1:-1]))

    control_point_positions = _optimize_control_point_positions(
        edge_to_control_points, expanded_node_positions, control_point_positions,
        origin, scale, k, initial_temperature, total_iterations, node_size,
        bundle_parallel_edges=False)

    edge_to_path = _get_path_through_control_points(
        edge_to_control_points, node_positions, control_point_positions)

    edge_to_path = _smooth_edge_paths(edge_to_path)

    return edge_to_path


def _initialize_selfloop_control_points(edges):
    """Represent each edge with string of control points."""
    edge_to_control_points = dict()
    for edge in edges:
        edge_to_control_points[edge] = [uuid4() for _ in range(5)]
    return edge_to_control_points


def _initialize_selfloop_control_point_positions(edge_to_control_points, node_positions, selfloop_radius, selfloop_angle):
    """Initialise the positions on a circle next to the node."""

    control_point_positions = dict()
    for edge, control_points in edge_to_control_points.items():
        positions = _get_selfloop_path(
            node_positions[edge[0]], selfloop_radius[edge], selfloop_angle[edge], len(control_points))
        control_point_positions.update(zip(control_points, positions))

    return control_point_positions


def _optimize_control_point_positions(
        edge_to_control_points, node_positions, control_point_positions,
        origin, scale, k, initial_temperature, total_iterations, node_size,
        bundle_parallel_edges):
    """Optimise the position of control points using the FR algorithm."""
    nodes = list(node_positions.keys())
    expanded_edges = _expand_edges(edge_to_control_points)
    expanded_node_positions = control_point_positions.copy() # TODO: may need deepcopy here
    expanded_node_positions.update(node_positions)

    if isinstance(node_size, float):
        node_size = {node : node_size for node in node_positions}

    # increase size of nodes so that there is a bit more clearance between edges and nodes
    node_size = {node : 2 * size for node, size in node_size.items()}

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)

        if bundle_parallel_edges:
            # Edge control points are repulsed by nodes but not by other edge control points.
            expanded_node_positions = get_fruchterman_reingold_layout.__wrapped__(
                expanded_edges,
                origin              = origin,
                scale               = scale,
                k                   = k,
                initial_temperature = initial_temperature,
                total_iterations    = total_iterations,
                node_size           = node_size,
                node_positions      = expanded_node_positions,
                fixed_nodes         = nodes,
                validate_positions  = False,
                get_repulsion       = _get_fr_repulsion_variant,
            )
        else:
            # Edge control points are repulsed by other edge control points.
            # This results in a separation of parallel edges.
            expanded_node_positions = get_fruchterman_reingold_layout.__wrapped__(
                expanded_edges,
                origin              = origin,
                scale               = scale,
                k                   = k,
                initial_temperature = initial_temperature,
                total_iterations    = total_iterations,
                node_size           = node_size,
                node_positions      = expanded_node_positions,
                fixed_nodes         = nodes,
                validate_positions  = False,
            )

    return {node : xy for node, xy in expanded_node_positions.items() if node not in nodes}


def _expand_edges(edge_to_control_points):
    """Create a new, expanded edge list, in which each edge is split into multiple segments.
    There are total_control_points + 1 segments / edges for each original edge.
    """
    expanded_edges = []
    for (source, target), control_points in edge_to_control_points.items():
        sources = [source] + control_points
        targets = control_points + [target]
        expanded_edges.extend(zip(sources, targets))
    return expanded_edges


def _get_fr_repulsion_variant(distance, direction, k):
    """Compute repulsive forces.

    This is a variant of the implementation in the original FR
    algorithm, in as much as repulsion only acts between fixed nodes
    and mobile nodes, not between mobile nodes and other mobile nodes.
    """
    total_mobile = distance.shape[1]
    distance = distance[total_mobile:]
    direction = direction[total_mobile:]
    magnitude = k**2 / distance
    vectors = direction * magnitude[..., None]
    return np.sum(vectors, axis=0)


def _get_path_through_control_points(edge_to_control_points, node_positions, control_point_positions):
    """Map each edge to an array of (optimised) control points positions."""
    edge_to_path = dict()
    for (source, target), control_points in edge_to_control_points.items():
        path = [node_positions[source]] \
            + [control_point_positions[node] for node in control_points] \
            + [node_positions[target]]
        edge_to_path[(source, target)] = np.array(path)
    return edge_to_path


def _smooth_edge_paths(edge_to_path, *args, **kwargs):
    """Fit splines through edge paths for smoother edge routing."""
    return {edge : _bspline(path, *args, **kwargs) for edge, path in edge_to_path.items()}


def get_arced_edge_paths(edges, node_positions, rad=1., selfloop_radius=0.1, selfloop_angle=np.pi/2):
    """Determine the edge layout, where edges are represented by arcs
    connecting the source and target node.

    Creates simple quadratic Bezier curves between nodes. The curves
    are created so that the middle control points (C1) are located at
    the same distance from the start (C0) and end points (C2) and the
    distance of the C1 to the line connecting C0-C2 is rad times the
    distance of C0-C2.

    Arguments:
    ----------
    edges : list of (source node ID, target node ID) 2-tuples
        The edges.
    node_positions : dict node ID : (x, y) positions
        The node positions.
    rad : float (default 1.0)
        The curvature of the arc.
    selfloop_radius : dict or float, default 0.1
        Dictionary mapping each self-loop edge to a radius. If float, all self-loops have the same radius.
    selfloop_angle : dict or float, default np.pi/2
        The starting angle of the self-loop in radians.

    Returns:
    --------
    edge_paths : dict edge : ndarray
        Dictionary mapping each edge to a list of edge segments.

    """
    edge_paths = dict()

    nonloops = [(source, target) for (source, target) in edges if source != target]
    nonloop_edge_paths = _get_arced_nonloop_edge_paths(nonloops, node_positions, rad)
    edge_paths.update(nonloop_edge_paths)

    selfloops = [(source, target) for (source, target) in edges if source == target]
    if selfloops:
        selfloop_radius = _normalize_numeric_argument(selfloop_radius, selfloops, 'selfloop_radius')
        selfloop_angle = _normalize_numeric_argument(selfloop_angle, selfloops, 'angle', allow_none=False)
        selfloop_edge_paths = _get_arced_selfloop_edge_paths(
            selfloops, node_positions, selfloop_radius, selfloop_angle)
        edge_paths.update(selfloop_edge_paths)

    return edge_paths


def _get_arced_nonloop_edge_paths(edges, node_positions, rad):
    edge_paths = dict()
    for source, target in edges:
        arc_factory = ConnectionStyle.Arc3(rad=rad)
        path = arc_factory(
            node_positions[source],
            node_positions[target],
            shrinkA=0., shrinkB=0.
            )
        edge_paths[(source, target)] = _bspline(path.vertices, 100)
    return edge_paths


def _get_arced_selfloop_edge_paths(edges, node_positions, selfloop_radius, selfloop_angle):
    return _get_straight_selfloop_edge_paths(edges, node_positions, selfloop_radius, selfloop_angle)


@_handle_multiple_components
@profile
def get_bundled_edge_paths(edges, node_positions,
                           k                       = 500.,
                           compatibility_threshold = 0.05,
                           total_cycles            = 6,
                           total_iterations        = 50,
                           step_size               = 0.04,
                           straighten_by           = 0.,
                           verbose                 = True,
):
    """Edge routing with bundled edge paths.

    Uses the FDEB algorithm as proposed in [Holten2009]_.
    This implementation follows the paper closely with the exception
    that instead of doubling the number of control point on each
    iteration (2n), a new control point is inserted between each
    existing pair of control points (2(n-1)+1), as proposed e.g. in Wu
    et al. (2015) [Wu2015]_.

    Parameters
    ----------

    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.
    k : float, default 500.
        The stiffness of the springs that connect control points.
    compatibility_threshold : float, default 0.05
        Edge pairs with a lower compatibility score are not bundled together.
        Set to zero to bundle all edges with each other regardless of compatibility.
        Set to one to prevent bundling of any (non-identical) edges.
    total_cycles : int, default 6
        The number of cycles. The number of control points (P) is doubled each cycle.
    total_iterations : int, default 50
        Number of iterations (I) in the first cycle. Iterations are reduced by 1/3 with each cycle.
    step_size : float, default 0.04
        Maximum step size (S) in the first cycle. Step sizes are halved each cycle.
    straighten_by : float, default 0.
        The amount of edge straightening applied after bundling.
        A small amount of straightening can help indicating the number of
        edges comprising a bundle by widening the bundle.
        If set to one, edges are fully un-bundled and plotted as stright lines.

    Returns
    -------
    edge_paths : dict
        Dictionary mapping each edge to an array of (x, y) coordinates representing its path.

    References
    ----------
    .. [Holten2009] Holten D and Van Wijk JJ. (2009) ‘Force-Directed edge
       bundling for graph visualization’, Computer Graphics Forum.

    .. [Wu2015] Wu J, Yu L, Yu H (2015) ‘Texture-based edge bundling: A
       web-based approach for interactively visualizing large graphs’,
       IEEE International Conference on Big Data.

    """

    # Filter out self-loops.
    if np.any([source == target for source, target in edges]):
        warnings.warn('Edge-bundling of self-loops not supported. Self-loops are removed from the edge list.')
        edges = [(source, target) for (source, target) in edges if source != target]

    # Filter out bi-directional edges.
    unidirectional_edges = set()
    for (source, target) in edges:
        if (target, source) not in unidirectional_edges:
            unidirectional_edges.add((source, target))
    reverse_edges = list(set(edges) - unidirectional_edges)
    edges = list(unidirectional_edges)

    edge_to_k = _get_k(edges, node_positions, k)

    edge_compatibility = _get_edge_compatibility(edges, node_positions, compatibility_threshold, verbose=verbose)

    edge_to_control_points = _initialize_bundled_control_points(edges, node_positions)

    for ii in range(total_cycles):
        edge_to_control_points = _expand_control_points(edge_to_control_points)

        for jj in range(total_iterations):
            if verbose: _print_progress_bar(ii * total_iterations + jj + 1, total_cycles * total_iterations, prefix="Bundling edges:")
            F = _get_Fs(edge_to_control_points, edge_to_k)
            F = _get_Fe(edge_to_control_points, edge_compatibility, F)
            edge_to_control_points = _update_control_point_positions(
                edge_to_control_points, F, step_size)

        step_size /= 2.
        total_iterations = int(2/3 * total_iterations)

    if straighten_by > 0.:
        edge_to_control_points = _straighten_edges(edge_to_control_points, straighten_by)

    edge_to_control_points = _smooth_edges(edge_to_control_points)

    # Add previously removed bi-directional edges back in.
    for (source, target) in reverse_edges:
        edge_to_control_points[(source, target)] = edge_to_control_points[(target, source)][::-1]

    return edge_to_control_points


def _get_k(edges, node_positions, k):
    """Assign each edge a stiffness depending on its length and the global stiffness constant."""
    return {(s, t) : k / np.linalg.norm(node_positions[t] - node_positions[s]) for (s, t) in edges}


@profile
def _get_edge_compatibility(edges, node_positions, threshold, verbose):
    """Compute the compatibility between all edge pairs."""
    # precompute edge segments, segment lengths and corresponding vectors
    edge_to_segment = {edge : Segment(node_positions[edge[0]], node_positions[edge[1]]) for edge in edges}

    edge_compatibility = list()
    for ii, (e1, e2) in enumerate(itertools.combinations(edges, 2)):
        if verbose: _print_progress_bar(ii + 1, len(edges) * (len(edges)-1) / 2, prefix="Initializing edge bundling:")
        P = edge_to_segment[e1]
        Q = edge_to_segment[e2]

        compatibility = 1
        compatibility *= _get_scale_compatibility(P, Q)
        if compatibility < threshold:
            continue # with next edge pair
        compatibility *= _get_position_compatibility(P, Q)
        if compatibility < threshold:
            continue # with next edge pair
        compatibility *= _get_angle_compatibility(P, Q)
        if compatibility < threshold:
            continue # with next edge pair
        compatibility *= _get_visibility_compatibility(P, Q)
        if compatibility < threshold:
            continue # with next edge pair

        # Also determine if one of the edges needs to be reversed:
        reverse = min(np.linalg.norm(P[0] - Q[0]), np.linalg.norm(P[1] - Q[1])) > \
            min(np.linalg.norm(P[0] - Q[1]), np.linalg.norm(P[1] - Q[0]))

        edge_compatibility.append((e1, e2, compatibility, reverse))

    return edge_compatibility


class Segment(object):
    def __init__(self, p0, p1):
        self.p0 = p0
        self.p1 = p1
        self.vector = p1 - p0
        self.length = np.linalg.norm(self.vector)
        self.unit_vector = self.vector / self.length
        self.midpoint = self.p0 * 0.5 * self.vector

    def __getitem__(self, idx):
        if idx == 0:
            return self.p0
        elif (idx == 1) or (idx == -1):
            return self.p1
        else:
            raise IndexError

    def get_orthogonal_projection_onto_segment(self, point):
        # Adapted from https://stackoverflow.com/a/61343727/2912349
        # The line extending the segment is parameterized as p0 + t (p1 - p0).
        # The projection falls where t = [(point-p0) . (p1-p0)] / |p1-p0|^2
        t = np.sum((point - self.p0) * self.vector) / self.length**2
        return self.p0 + t * self.vector

#     def get_interior_angle_with(self, other_segment):
#         # Adapted from: https://stackoverflow.com/a/13849249/2912349
#         return np.arccos(np.clip(np.dot(self.unit_vector, other_segment.unit_vector), -1.0, 1.0))


# def _get_angle_compatibility(P, Q):
#     return np.abs(np.cos(P.get_interior_angle_with(Q)))


def _get_angle_compatibility(P, Q):
    """Compute the angle compatibility between two segments P and Q.
    The angle compatibility is high if the interior angle between them is small.

    """
    return np.abs(np.clip(np.dot(P.unit_vector, Q.unit_vector), -1.0, 1.0))


def _get_scale_compatibility(P, Q):
    """Compute the scale compatibility between two segments P and Q.
    The scale compatibility is high if their lengths are similar.

    """
    avg = 0.5 * (P.length + Q.length)

    # The definition in the paper is rubbish, as the result is not on the interval [0, 1].
    # For example, consider an two edges, both 0.5 long:
    # return 2 / (avg * min(length_P, length_Q) + max(length_P, length_Q) / avg)

    # my original alternative:
    # return min(length_P/length_Q, length_Q/length_P)

    # typo in original paper corrected in Graser et al. (2019)
    return 2 / (avg / min(P.length, Q.length) + max(P.length, Q.length) / avg)


def _get_position_compatibility(P, Q):
    """Compute the position compatibility between two segments P and Q.
    The position compatibility is high if the distance between their midpoints is small.

    """
    avg = 0.5 * (P.length + Q.length)
    distance_between_midpoints = np.linalg.norm(Q.midpoint - P.midpoint)
    # This is the definition from the paper, but the scaling should probably be more aggressive.
    return avg / (avg + distance_between_midpoints)


def _get_visibility_compatibility(P, Q):
    """Compute the visibility compatibility between two segments P and Q.
    The visibility compatibility is low if bundling would occlude any of the end points of the segments.

    """
    return min(_get_visibility(P, Q), _get_visibility(Q, P))


@profile
def _get_visibility(P, Q):
    I0 = P.get_orthogonal_projection_onto_segment(Q[0])
    I1 = P.get_orthogonal_projection_onto_segment(Q[1])
    I = Segment(I0, I1)
    distance_between_midpoints = np.linalg.norm(P.midpoint - I.midpoint)
    visibility = 1 - 2 * distance_between_midpoints / I.length
    return max(visibility, 0)


def _initialize_bundled_control_points(edges, node_positions):
    """Initialise each edge with two control points, the positions of the source and target nodes."""
    edge_to_control_points = dict()
    for source, target in edges:
        edge_to_control_points[(source, target)] \
            = np.array([node_positions[source], node_positions[target]])
    return edge_to_control_points


def _expand_control_points(edge_to_control_points):
    "Place a new control point between each pair of existing control points."
    for edge, control_points_old in edge_to_control_points.items():
        total_control_points_old = len(control_points_old)
        total_control_points_new = 2 * (total_control_points_old - 1) + 1
        control_points_new = np.zeros((total_control_points_new, 2))
        for ii in range(total_control_points_new):
            if (ii+1) % 2: # ii is even
                control_points_new[ii] = control_points_old[int(ii/2)]
            else: # ii is odd
                p1 = control_points_old[int((ii-1)/2)]
                p2 = control_points_old[int((ii+1)/2)]
                control_points_new[ii] = 0.5 * (p2 - p1) + p1
        edge_to_control_points[edge] = control_points_new
    return edge_to_control_points


def _get_Fs(edge_to_control_points, k):
    """Compute all spring forces."""
    out = dict()
    for edge, control_points in edge_to_control_points.items():
        delta = np.zeros_like(control_points)
        diff = np.diff(control_points, axis=0)
        delta[1:-1] -= diff[:-1]
        delta[1:-1] += diff[1:]
        kp = k[edge] / (len(control_points) - 1)
        out[edge] = kp * delta
    return out


@profile
def _get_Fe(edge_to_control_points, edge_compatibility, out):
    """Compute all electrostatic forces."""
    for e1, e2, compatibility, reverse in edge_compatibility:
        P = edge_to_control_points[e1]
        Q = edge_to_control_points[e2]

        if not reverse:
            # i.e. if source/source or target/target closest
            delta = Q - P
        else:
            # need to reverse one set of control points
            delta = Q[::-1] - P

        # # desired computation:
        # distance = np.linalg.norm(delta, axis=1)
        # displacement = compatibility * delta / distance[..., None]**2

        # actually much faster:
        distance_squared = delta[:, 0]**2 + delta[:, 1]**2
        displacement = compatibility * delta / distance_squared[..., None]

        # Don't move the first and last control point, which are just the node positions.
        displacement[0] = 0
        displacement[-1] = 0

        out[e1] += displacement
        if not reverse:
            out[e2] -= displacement
        else:
            out[e2] -= displacement[::-1]

    return out


def _update_control_point_positions(edge_to_control_points, F, step_size):
    """Update control point positions using the calculated net forces."""
    for edge, displacement in F.items():
        displacement_length = np.clip(np.linalg.norm(displacement), 1e-12, None) # prevent divide by 0 error in next line
        displacement = displacement / displacement_length * np.clip(displacement_length, None, step_size)
        edge_to_control_points[edge] += displacement
    return edge_to_control_points


def _smooth_edges(edge_to_path):
    """Wraps _smooth_path()."""
    return {edge : _smooth_path(path) for edge, path in edge_to_path.items()}


def _smooth_path(path):
    """Smooth a path by fitting a univariate spline.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/52020098/2912349

    """

    # Compute the linear length along the line:
    distance = np.cumsum( np.sqrt(np.sum( np.diff(path, axis=0)**2, axis=1 )) )
    distance = np.insert(distance, 0, 0)/distance[-1]

    # Compute a spline function for each dimension:
    splines = [UnivariateSpline(distance, coords, k=3, s=.001) for coords in path.T]

    # Computed the smoothed path:
    alpha = np.linspace(0, 1, 100)
    return np.vstack([spl(alpha) for spl in splines]).T


def _straighten_edges(edge_to_path, straighten_by):
    """Wraps _straigthen_path()"""
    return {edge : _straighten_path(path, straighten_by) for edge, path in edge_to_path.items()}


def _straighten_path(path, straighten_by):
    """Straigthen a path by computing the weighted average between the path and
    a straight line connecting the end points.

    """
    p0 = path[0]
    p1 = path[-1]
    n = len(path)
    return (1 - straighten_by) * path \
        + straighten_by * (p0 + np.linspace(0, 1, n)[:, np.newaxis] * (p1 - p0))


def _shift_edge(x1, y1, x2, y2, delta):
    """Determine the parallel to a segment defined by points p1: (x1, y1) and p2 : (x2, y2) at a distance delta."""
    # convert segment into a vector
    v = np.r_[x2-x1, y2-y1]
    # compute orthogonal vector
    v = np.r_[-v[1], v[0]]
    # convert to orthogonal unit vector
    v = v / np.linalg.norm(v)
    # compute offsets
    dx, dy = delta * v
    # return new coordinates of point p1' and p2'
    return x1+dx, y1+dy, x2+dx, y2+dy


class StraightEdgeLayout(object):

    def __init__(self, edges, node_positions, selfloop_radius=0.1, selfloop_angle=None):
        self.node_positions = node_positions
        self.edges = edges
        self.nonloops = [(source, target) for (source, target) in edges if source != target]
        self.selfloops = [(source, target) for (source, target) in edges if source == target]

        if self.selfloops:
            self.selfloop_radius = _normalize_numeric_argument(selfloop_radius, self.selfloops, 'selfloop_radius')
            if selfloop_angle is not None: # can be zero!
                self.selfloop_angle = _normalize_numeric_argument(selfloop_angle, self.selfloops, 'selfloop_angle', allow_none=True)
            elif (selfloop_angle is None) and (not self.nonloops):
                self.selfloop_angle = _normalize_numeric_argument(np.pi/2, self.selfloops, 'selfloop_angle')

        self.edge_paths = dict()
        self.nonloop_edge_paths = dict()
        self.selfloop_edge_paths = dict()


    def get(self, stale_nodes=None, approximate=False):
        if stale_nodes is not None: # all
            stale_nodes = self.node_positions
            stale_nonloops = [(source, target) for source, target in self.nonloops \
                              if (source in stale_nodes) or (target in stale_nodes)]
            stale_selfloops = [(source, target) for source, target in self.selfloops \
                               if (source in stale_nodes) or (target in stale_nodes)]
        else:
            stale_nonloops = self.nonloops
            stale_selfloops = self.selfloops

        new_edge_paths = dict()
        if stale_nonloops:
            if not approximate:
                new_nonloop_edge_paths = self.get_nonloop_edge_paths(stale_nonloops)
            else:
                new_nonloop_edge_paths = self.approximate_nonloop_edge_paths(stale_nonloops)
            new_edge_paths.update(new_nonloop_edge_paths)
            self.nonloop_edge_paths.update(new_nonloop_edge_paths)
            self.edge_paths.update(new_nonloop_edge_paths)

        if stale_selfloops:
            if not approximate:
                new_selfloop_edge_paths = self.get_selfloop_edge_paths(stale_selfloops)
            else:
                new_selfloop_edge_paths = self.approximate_selfloop_edge_paths(stale_selfloops)
            new_edge_paths.update(new_selfloop_edge_paths)
            self.selfloop_edge_paths.update(new_selfloop_edge_paths)
            self.edge_paths.update(new_selfloop_edge_paths)

        return new_edge_paths


    def get_nonloop_edge_paths(self, edges):
        return _get_straight_nonloop_edge_paths(
            edges, self.node_positions)


    def approximate_nonloop_edge_paths(self, edges):
        return _get_straight_nonloop_edge_paths(
            edges, self.node_positions)


    def get_selfloop_edge_paths(self, edges):
        if hasattr(self, "selfloop_angle"):
            selfloop_angle = self.selfloop_angle
            if None in selfloop_angle.values():
                selfloop_angle.update(
                    _get_optimal_selfloop_angles(
                        [edge for edge, angle in selfloop_angle.items() if isinstance(angle, type(None))],
                        self.selfloop_radius,
                        self.node_positions,
                        self.nonloop_edge_paths,
                    )
                )
        elif self.nonloops:
            selfloop_angle = _get_optimal_selfloop_angles(
                self.selfloops, self.selfloop_radius, self.node_positions, self.nonloop_edge_paths)
        else:
            selfloop_angle = _get_decent_selfloop_angles(
                edges, self.node_positions)
        return _get_straight_selfloop_edge_paths(
            edges, self.node_positions, self.selfloop_radius, selfloop_angle)


    def approximate_selfloop_edge_paths(self, edges):
        if hasattr(self, "selfloop_angle"):
            selfloop_angle = self.selfloop_angle
            if None in selfloop_angle.values():
                selfloop_angle.update(
                    _get_decent_selfloop_angles(
                        [edge for edge, angle in selfloop_angle.items() if isinstance(angle, type(None))],
                        self.node_positions,
                    )
                )
        else:
            selfloop_angle = _get_decent_selfloop_angles(
                edges, self.node_positions)
        return _get_straight_selfloop_edge_paths(
            edges, self.node_positions, self.selfloop_radius, selfloop_angle)


    def add_node(self, node, position):
        # self.node_positions[node] = position
        pass


    def delete_node(self, node):
        # del self.node_positions[node]
        pass


    def add_edge(self, edge):
        source, target = edge
        if source == target:
            self.selfloops.append(edge)
            if self.selfloop_radius:
                self.selfloop_radius[edge] = np.median(list(self.selfloop_radius.values()))
            else:
                self.selfloop_radius[edge] = 0.1
        else:
            self.nonloops.append(edge)
        # self.edges.append(edge)


    def delete_edge(self, edge):
        source, target = edge[:2]
        if source == target:
            self.selfloops.remove(edge)
            del self.selfloop_radius[edge]
            del self.selfloop_edge_paths[edge]
            del self.edge_paths[edge]
        else:
            self.nonloops.remove(edge)
            del self.nonloop_edge_paths[edge]
            del self.edge_paths[edge]
        # self.edges.remove(edge)


class ArcEdgeLayout(StraightEdgeLayout):

    def __init__(self, edges, node_positions, selfloop_radius=0.1, selfloop_angle=np.pi/2, rad=1.):
        self.rad = rad
        super().__init__(
            edges,
            node_positions,
            selfloop_radius,
            selfloop_angle,
        )


    def get_nonloop_edge_paths(self, edges):
        return _get_arced_nonloop_edge_paths(edges, self.node_positions, self.rad)


    def approximate_nonloop_edge_paths(self, edges):
        return _get_arced_nonloop_edge_paths(edges, self.node_positions, self.rad)


    def get_selfloop_edge_paths(self, edges):
        if hasattr(self, "selfloop_angle"):
            selfloop_angle = self.selfloop_angle
            if None in selfloop_angle.values():
                selfloop_angle.update(
                    _get_optimal_selfloop_angles(
                        [edge for edge, angle in selfloop_angle.items() if isinstance(angle, type(None))],
                        self.selfloop_radius,
                        self.node_positions,
                        self.nonloop_edge_paths,
                    )
                )
        elif self.nonloops:
            selfloop_angle = _get_optimal_selfloop_angles(
                self.selfloops,
                self.selfloop_radius,
                self.node_positions,
                self.nonloop_edge_paths
            )
        else:
            selfloop_angle = _get_decent_selfloop_angles(
                edges, self.node_positions)
        return  _get_arced_selfloop_edge_paths(
            edges,
            self.node_positions,
            self.selfloop_radius,
            selfloop_angle,
        )


    def approximate_selfloop_edge_paths(self, edges):
        if hasattr(self, "selfloop_angle"):
            selfloop_angle = self.selfloop_angle
            if None in selfloop_angle.values():
                selfloop_angle.update(
                    _get_decent_selfloop_angles(
                        [edge for edge, angle in selfloop_angle.items() if isinstance(angle, type(None))],
                        self.node_positions,
                    )
                )
        else:
            selfloop_angle = _get_decent_selfloop_angles(
                edges, self.node_positions)
        return  _get_arced_selfloop_edge_paths(
            edges,
            self.node_positions,
            self.selfloop_radius,
            selfloop_angle,
        )


class ArcDiagramEdgeLayout(ArcEdgeLayout):

    def __init__(self, edges, node_positions, selfloop_radius=0.1, selfloop_angle=np.pi/2, rad=1., above=True):
        self.above = above
        super().__init__(
            edges, node_positions,
            selfloop_radius=selfloop_radius,
            selfloop_angle=selfloop_angle,
            rad=rad
        )

    def get_nonloop_edge_paths(self, edges):
        edge_paths = super().get_nonloop_edge_paths(edges)
        edge_paths = _lateralize_arced_edge_paths(edge_paths, self.node_positions, self.above)
        return edge_paths


    def approximate_nonloop_edge_paths(self, edges):
        edge_paths = super().approximate_nonloop_edge_paths(edges)
        edge_paths = _lateralize_arced_edge_paths(edge_paths, self.node_positions, self.above)
        return edge_paths


    def get_selfloop_edge_paths(self, edges):
        edge_paths = super().get_selfloop_edge_paths(edges)
        edge_paths = _lateralize_arced_edge_paths(edge_paths, self.node_positions, self.above)
        return edge_paths


    def approximate_selfloop_edge_paths(self, edges):
        edge_paths = super().approximate_selfloop_edge_paths(edges)
        edge_paths = _lateralize_arced_edge_paths(edge_paths, self.node_positions, self.above)
        return edge_paths


def _lateralize_arced_edge_paths(edge_paths, node_positions, above):
    """Ensure that edge paths are all either above or below the line passing through nodes."""
    for (source, target), path in edge_paths.items():
        edge_paths[(source, target)] = \
            _lateralize(path, node_positions[source], node_positions[target], above)
    return edge_paths


def _lateralize(path, p1, p2, above):
    if not np.all(np.isclose(p1, p2)):
        gradient, intercept = _get_gradient_and_intercept(p1, p2)
    else:
        gradient, intercept = 0., p1[1]
    mask = _is_above_line(path, gradient, intercept)
    if above:
        mask = np.invert(mask)
    path[mask] = _reflect_across_line(path[mask], gradient, intercept)
    return path


class CurvedEdgeLayout(StraightEdgeLayout):

    def __init__(self, edges, node_positions,
                 selfloop_radius       = 0.1,
                 selfloop_angle        = None,
                 origin                = np.array([0, 0]),
                 scale                 = np.array([1, 1]),
                 k                     = 0.1,
                 initial_temperature   = 0.01,
                 total_iterations      = 50,
                 node_size             = 0.,
                 bundle_parallel_edges = True):

        super().__init__(edges, node_positions, selfloop_radius, selfloop_angle)

        self.origin                = origin
        self.scale                 = scale
        self.k                     = k
        self.initial_temperature   = initial_temperature
        self.total_iterations      = total_iterations
        self.node_size             = node_size
        self.bundle_parallel_edges = bundle_parallel_edges

        if self.bundle_parallel_edges:
            self.parallel_edges = []
            other_edges = []
            for (source, target) in self.nonloops:
                if (target, source) in self.nonloops:
                    if (target, source) in self.parallel_edges:
                        pass
                    else:
                        self.parallel_edges.append((source, target))
                else:
                    other_edges.append((source, target))
            self.nonloops = self.parallel_edges + other_edges


    def get_nonloop_edge_paths(self, edges):
        edge_paths =  _get_curved_nonloop_edge_paths(edges, self.node_positions,
                self.origin, self.scale, self.k, self.initial_temperature,
                self.total_iterations, self.node_size, self.bundle_parallel_edges)
        if self.bundle_parallel_edges:
            for (source, target) in set(self.parallel_edges).intersection(set(edges)):
                edge_paths[(target, source)] = edge_paths[(source, target)][::-1]
        return edge_paths


    def approximate_nonloop_edge_paths(self, edges):
        new_nonloop_edge_paths = super().approximate_nonloop_edge_paths(edges)
        if self.bundle_parallel_edges:
            for (source, target) in set(self.parallel_edges).intersection(set(edges)):
                new_nonloop_edge_paths[(target, source)] = new_nonloop_edge_paths[(source, target)][::-1]
        return new_nonloop_edge_paths


    def get_selfloop_edge_paths(self, edges):
        if hasattr(self, "selfloop_angle"):
            selfloop_angle = self.selfloop_angle
            if None in selfloop_angle.values():
                selfloop_angle.update(
                    _get_optimal_selfloop_angles(
                        [edge for edge, angle in selfloop_angle.items() if isinstance(angle, type(None))],
                        self.selfloop_radius,
                        self.node_positions,
                        self.nonloop_edge_paths)
                )
        elif self.nonloops:
            selfloop_angle = _get_optimal_selfloop_angles(
                self.selfloops, self.selfloop_radius, self.node_positions, self.nonloop_edge_paths)
        else:
            selfloop_angle = _get_decent_selfloop_angles(
                edges, self.node_positions)

        return  _get_curved_selfloop_edge_paths(
            edges, self.node_positions, self.selfloop_radius, selfloop_angle,
            self.origin, self.scale, self.k, self.initial_temperature,
            self.total_iterations, self.node_size, self.nonloop_edge_paths)


class BundledEdgeLayout(StraightEdgeLayout):

    def __init__(self, edges, node_positions,
                 selfloop_radius         = 0.1,
                 selfloop_angle          = None,
                 k                       = 500.,
                 compatibility_threshold = 0.05,
                 total_cycles            = 6,
                 total_iterations        = 50,
                 step_size               = 0.04,
                 straighten_by           = 0.,
                 verbose                 = True,
                 ):

        super().__init__(edges, node_positions, selfloop_radius, selfloop_angle)

        self.k                       = k
        self.compatibility_threshold = compatibility_threshold
        self.total_cycles            = total_cycles
        self.total_iterations        = total_iterations
        self.step_size               = step_size
        self.straighten_by           = straighten_by
        self.verbose                 = verbose


    def get_nonloop_edge_paths(self, edges):
        return get_bundled_edge_paths(edges, self.node_positions,
                                      self.k,
                                      self.compatibility_threshold,
                                      self.total_cycles,
                                      self.total_iterations,
                                      self.step_size,
                                      self.straighten_by,
                                      self.verbose)


def _expand_nonloop_edge_paths(edge_paths, edge_to_ids, edge_width):
    expanded_edge_paths = dict()
    for (source, target), base_path in edge_paths.items():
        edge_ids = edge_to_ids[(source, target)]

        # pre-compute offsets
        offsets = np.cumsum([0] + [3 * edge_width[(source, target, eid)] for eid in edge_ids[:-1]], dtype=float)
        if (target, source) in edge_to_ids:
            # offset edges further as there are edges running in the reverse direction
            offsets += 1.5 * edge_width[(source, target, edge_ids[0])]
        else: # center edges w.r.t. node as there are no reverse edges
            offsets -= np.mean(offsets)

        base_path = edge_paths[(source, target)]
        tail_position = base_path[0]
        head_position = base_path[-1]

        length = np.linalg.norm(head_position - tail_position)
        delta = 0.05 * length
        for ii, eid in enumerate(edge_ids):
            # offset
            path = _get_parallel_line(base_path, offsets[ii])
            # shorten
            path = _shorten_spline_by(path, delta)
            path = _shorten_spline_by(path[::-1], delta)[::-1]
            # converge on source and target nodes
            path = np.vstack([tail_position, path, head_position])
            # resample -- otherwise, a bug occurs where short paths loose their offset
            path = _resample_spline(path, max(len(base_path), 20))
            # assign
            expanded_edge_paths[(source, target, eid)] = path

    return expanded_edge_paths


def _expand_selfloop_edge_paths(edge_paths, edge_to_ids, edge_width, selfloop_radius):
    expanded_edge_paths = dict()
    for (source, target), base_path in edge_paths.items():
        edge_ids = edge_to_ids[(source, target)]
        # pre-compute offsets
        offsets = np.cumsum([0] + [3 * edge_width[(source, target, eid)] for eid in edge_ids[:-1]], dtype=float)
        if (target, source) in edge_to_ids:
            # offset edges further as there are edges running in the reverse direction
            offsets += 1.5 * edge_width[(source, target, edge_ids[0])]
        else: # center edges w.r.t. node as there are no reverse edges
            offsets -= np.mean(offsets)

        base_path = edge_paths[(source, target)]
        tail_position = base_path[0]
        head_position = base_path[-1]

        length = 2 * np.pi * selfloop_radius[(source, target)]
        delta = 0.1 * length
        for ii, eid in enumerate(edge_ids):
            # offset
            path = _get_parallel_line(base_path, -offsets[ii])
            # shorten
            path = _shorten_spline_by(path, delta)
            path = _shorten_spline_by(path[::-1], delta)[::-1]
            # converge on source and target nodes
            path = np.vstack([tail_position, path, head_position])
            # smooth
            path = _smooth_path(path)
            # assign
            expanded_edge_paths[(source, target, eid)] = path

    return expanded_edge_paths


class _MultiGraphEdgeLayout(object):


    def __init__(self, edges, node_positions, edge_width, *args, **kwargs):
        self.edge_width = edge_width
        self.edge_to_ids = _map_multigraph_edges_to_ids(edges)
        super().__init__(list(self.edge_to_ids.keys()), node_positions, *args, **kwargs)


    def get_nonloop_edge_paths(self, *args, **kwargs):
        edge_paths = super().get_nonloop_edge_paths(*args, **kwargs)
        return _expand_nonloop_edge_paths(edge_paths, self.edge_to_ids, self.edge_width)


    def approximate_nonloop_edge_paths(self, *args, **kwargs):
        edge_paths = super().approximate_nonloop_edge_paths(*args, **kwargs)
        return _expand_nonloop_edge_paths(edge_paths, self.edge_to_ids, self.edge_width)


    def get_selfloop_edge_paths(self, *args, **kwargs):
        edge_paths = super().get_selfloop_edge_paths(*args, **kwargs)
        return _expand_selfloop_edge_paths(edge_paths, self.edge_to_ids, self.edge_width, self.selfloop_radius)


    def approximate_selfloop_edge_paths(self, *args, **kwargs):
        edge_paths = super().approximate_selfloop_edge_paths(*args, **kwargs)
        return _expand_selfloop_edge_paths(edge_paths, self.edge_to_ids, self.edge_width, self.selfloop_radius)


    def add_edge(self, edge, width):
        source, target, key = edge
        self.edge_width[edge] = width
        if (source, target) in self.edge_to_ids:
            self.edge_to_ids[(source, target)].append(key)
        else:
            self.edge_to_ids[(source, target)] = [key]
        return super().add_edge(edge[:2])


    def delete_edge(self, edge):
        source, target, key = edge
        del self.edge_width[edge]
        self.edge_to_ids[(source, target)].remove(key)

        if source == target:
            del self.selfloop_edge_paths[edge]
            del self.edge_paths[edge]
        else:
            del self.nonloop_edge_paths[edge]
            del self.edge_paths[edge]

        if not self.edge_to_ids[(source, target)]:
            del self.edge_to_ids[(source, target)]
            if source == target:
                self.selfloops.remove((source, target))
                del self.selfloop_radius[(source, target)]
            else:
                self.nonloops.remove((source, target))
            # self.edges.remove(edge)


class MultiGraphStraightEdgeLayout(_MultiGraphEdgeLayout, StraightEdgeLayout):
    pass


class MultiGraphCurvedEdgeLayout(_MultiGraphEdgeLayout, CurvedEdgeLayout):
    pass


class MultiGraphArcDiagramEdgeLayout(_MultiGraphEdgeLayout, ArcDiagramEdgeLayout):
    pass


class MultiGraphBundledEdgeLayout(_MultiGraphEdgeLayout, BundledEdgeLayout):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Implements the BaseMultiGraph and MultiGraph classes.
"""

import matplotlib.pyplot as plt

from ._graph_classes import BaseGraph, Graph
from ._parser import parse_multigraph, _parse_multigraph_edge_list
from ._artists import EdgeArtist
from ._edge_layout import (
    MultiGraphStraightEdgeLayout,
    MultiGraphCurvedEdgeLayout,
    MultiGraphBundledEdgeLayout,
    MultiGraphArcDiagramEdgeLayout,
)
from ._utils import (
    _simplify_multigraph,
)


class BaseMultiGraph(BaseGraph):
    """The MultiGraph base class.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID, edge key) tuple.
    nodes : list or None, default None
        List of nodes. Required argument if any node in the graph is unconnected.
        If None, `nodes` is initialised to the set of the flattened `edges`.
    node_layout : str or dict, default 'spring'
        If a string, the node positions are computed using the indicated method:

        - 'random'       : place nodes in random positions;
        - 'circular'     : place nodes regularly spaced on a circle;
        - 'spring'       : place nodes using a force-directed layout (Fruchterman-Reingold algorithm);
        - 'dot'          : place nodes using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'radial'       : place nodes radially using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'community'    : place nodes such that nodes belonging to the same community are grouped together;
        - 'bipartite'    : place nodes regularly spaced on two parallel lines;
        - 'multipartite' : place nodes regularly spaced on several parallel lines;
        - 'shell'        : place nodes regularly spaced on concentric circles;
        - 'geometric'    : place nodes according to the length of the edges between them.

        If a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_random_layout`
        - :py:func:`get_circular_layout`
        - :py:func:`get_fruchterman_reingold_layout`
        - :py:func:`get_sugiyama_layout`
        - :py:func:`get_radial_tree_layout`
        - :py:func:`get_community_layout`
        - :py:func:`get_bipartite_layout`
        - :py:func:`get_multipartite_layout`
        - :py:func:`get_shell_layout`
        - :py:func:`get_geometric_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Edge color.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout : str or dict (default 'straight')
        If edge_layout is a string, determine the layout internally:

        - 'straight' : draw edges as straight lines
        - 'curved'   : draw edges as curved splines; the spline control points are optimised to avoid other nodes and edges
        - 'bundled'  : draw edges as edge bundles
        - 'arc'      : draw edges as arcs with a fixed curvature

        If edge_layout is a dict, the keys are edges and the values are edge paths
        in the form iterables of (x, y) tuples, the edge segments.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to edge layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_straight_edge_paths`
        - :py:func:`get_curved_edge_paths`
        - :py:func:`get_bundled_edge_paths`
        - :py:func:`get_arced_edge_paths`

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`BaseGraph`, :py:class:`MultiGraph`

    """

    def _initialize_edges(self, edges):
        return _parse_multigraph_edge_list(edges)


    def _get_node_positions(self, node_layout, node_layout_kwargs, edges, *args, **kwargs):
        return super()._get_node_positions(node_layout, node_layout_kwargs, _simplify_multigraph(edges), *args, **kwargs)


    def _initialize_edge_layout(self, edge_layout, edge_layout_kwargs):
        if edge_layout_kwargs is None:
            edge_layout_kwargs = dict()

        # TODO: improve handling of dict arguments such as selfloop_radius
        selfloops = [(source, target) for (source, target, eid) in self.edges if source==target]
        if selfloops:
            if 'selfloop_radius' in edge_layout_kwargs:
                edge_layout_kwargs['selfloop_radius'] = \
                    _normalize_numeric_argument(edge_layout_kwargs['selfloop_radius'], selfloops, 'selfloop_radius')
            else:
                edge_layout_kwargs['selfloop_radius'] = \
                    {(node, node) : 1.5 * self.node_size[node] for node, _ in selfloops}

        if isinstance(edge_layout, str):
            if edge_layout == "straight":
                edge_layout = MultiGraphStraightEdgeLayout(self.edges, self.node_positions, self.edge_width, **edge_layout_kwargs)
            elif edge_layout == "curved":
                edge_layout_kwargs.setdefault('node_size', self.node_size)
                edge_layout_kwargs.setdefault('origin', self.origin)
                edge_layout_kwargs.setdefault('scale', self.scale)
                edge_layout = MultiGraphCurvedEdgeLayout(self.edges, self.node_positions, self.edge_width, **edge_layout_kwargs)
            elif edge_layout == "bundled":
                edge_layout = MultiGraphBundledEdgeLayout(self.edges, self.node_positions, self.edge_width, **edge_layout_kwargs)
            elif edge_layout == "arc":
                edge_layout = MultiGraphArcDiagramEdgeLayout(self.edges, self.node_positions, self.edge_width, **edge_layout_kwargs)
            else:
                raise NotImplementedError(f"Variable edge_layout one of 'straight', 'curved', 'bundled', or 'arc', not {edge_layout}")
            edge_paths = edge_layout.get()

        elif isinstance(edge_layout, dict):
            _check_completeness(edge_layout, self.edges, 'edge_layout')
            edge_paths = edge_layout
            edge_layout = MultigraphStraightEdgeLayout(self.edges, self.node_positions, self.edge_width, **edge_layout_kwargs)
            edge_layout.edge_paths.update(edge_paths)

        else:
            raise TypeError("Variable `edge_layout` either a string or a dict mapping edges to edge paths.")

        return edge_layout


    def draw_edges(self, edges, edge_path, edge_width, edge_color, edge_alpha,
                   edge_zorder, arrows, node_artists):
        """Draw or update edge artists.

        Parameters
        ----------
        edges : list
            The edges of the graph, with each edge being represented by a
            (source node ID, target node ID, edge ID) tuple.
        edge_path : dict
            Mapping of edges to arrays of (x, y) tuples, the edge path coordinates.
        edge_width : dict
            Mapping of edges to floats, the edge widths.
        edge_color : dict
            Mapping of edges to valid matplotlib color specifications, the edge colors.
        edge_alpha : dict
            Mapping of edges to floats, the edge transparencies.
        edge_zorder : dict
            Mapping of edges to ints, the edge z-order values.
        arrows : bool
            If True, draw edges with arrow heads.
        node_artists : dict
            Mapping of nodes to node artists. Required to offset edges from nodes.

        Returns
        -------
        self.edge_artists: dict
            Updates mapping of edges to corresponding edge artists.

        """

        for edge in edges:

            curved = False if (len(edge_path[edge]) == 2) else True

            source, target, eid = edge

            if arrows:
                head_length = 2 * edge_width[edge]
                head_width = 3 * edge_width[edge]
            else:
                head_length = 0
                head_width = 0

            edge_artist = EdgeArtist(
                midline     = edge_path[edge],
                width       = edge_width[edge],
                facecolor   = edge_color[edge],
                alpha       = edge_alpha[edge],
                head_length = head_length,
                head_width  = head_width,
                edgecolor   = 'none',
                linewidth   = 0.,
                head_offset = node_artists[target].get_head_offset(edge_path[edge]),
                tail_offset = node_artists[source].get_tail_offset(edge_path[edge]),
                shape       = "full",
                curved      = curved,
                zorder      = edge_zorder[edge],
            )
            self.ax.add_patch(edge_artist)

            if edge in self.edge_artists:
                self.edge_artists[edge].remove()
            self.edge_artists[edge] = edge_artist


class MultiGraph(BaseMultiGraph, Graph):
    """Parses the given graph data object and initialises the :py:class:`BaseMultiGraph` object.

    If the given graph includes edge weights, then these are mapped to colors using the `edge_cmap` parameter.

    Parameters
    ----------
    graph: various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source node ID, target node ID, edge key) or
          (source node ID, target node ID, edge key, weight) tuples,
          or equivalent (E, 3) or (E, 4) ndarray (where E is the number of edges).
        - Adjacency matrix:
          A (V, V, L) ndarray, where V is the number of nodes/vertices, and L is the number of layers.
          The absence of a connection is indicated by a zero.
        - networkx.MultiGraph or igraph.Graph object

    node_layout : str or dict, default 'spring'
        If a string, the node positions are computed using the indicated method:

        - 'random'       : place nodes in random positions;
        - 'circular'     : place nodes regularly spaced on a circle;
        - 'spring'       : place nodes using a force-directed layout (Fruchterman-Reingold algorithm);
        - 'dot'          : place nodes using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'radial'       : place nodes radially using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'community'    : place nodes such that nodes belonging to the same community are grouped together;
        - 'bipartite'    : place nodes regularly spaced on two parallel lines;
        - 'multipartite' : place nodes regularly spaced on several parallel lines;
        - 'shell'        : place nodes regularly spaced on concentric circles;
        - 'geometric'    : place nodes according to the length of the edges between them.

        If a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_random_layout`
        - :py:func:`get_circular_layout`
        - :py:func:`get_fruchterman_reingold_layout`
        - :py:func:`get_sugiyama_layout`
        - :py:func:`get_radial_tree_layout`
        - :py:func:`get_community_layout`
        - :py:func:`get_bipartite_layout`
        - :py:func:`get_multipartite_layout`
        - :py:func:`get_shell_layout`
        - :py:func:`get_geometric_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Edge color.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout : str or dict (default 'straight')
        If edge_layout is a string, determine the layout internally:

        - 'straight' : draw edges as straight lines
        - 'curved'   : draw edges as curved splines; the spline control points are optimised to avoid other nodes and edges
        - 'bundled'  : draw edges as edge bundles
        - 'arc'      : draw edges as arcs with a fixed curvature

        If edge_layout is a dict, the keys are edges and the values are edge paths
        in the form iterables of (x, y) tuples, the edge segments.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to edge layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_straight_edge_paths`
        - :py:func:`get_curved_edge_paths`
        - :py:func:`get_bundled_edge_paths`
        - :py:func:`get_arced_edge_paths`

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`BaseMultiGraph`, :py:class:`Graph`, :py:class:`InteractiveMultiGraph`

    """

    def __init__(self, *args, **kwargs):
        Graph.__init__(self, *args, **kwargs)


    def _parse_input(self, graph):
        return parse_multigraph(graph)



"""
This module provides the LPCMCI class.

Classes:
    LPCMCI: class containing the LPCMCI causal discovery algorithm.
"""

from causalflow.causal_discovery.tigramite.lpcmci import LPCMCI as lpcmci
from tigramite.independence_tests.independence_tests_base import CondIndTest
import tigramite.data_processing as pp
import numpy as np
from causalflow.CPrinter import CPLevel, CP
from causalflow.basics.constants import *
from causalflow.graph.DAG import DAG
from causalflow.preprocessing.data import Data
from causalflow.causal_discovery.CausalDiscoveryMethod import CausalDiscoveryMethod 

class LPCMCI(CausalDiscoveryMethod):
    """LPCMCI causal discovery method."""
    
    def __init__(self, 
                 data: Data,
                 min_lag, max_lag, 
                 val_condtest: CondIndTest, 
                 verbosity: CPLevel,
                 alpha = 0.05, 
                 resfolder = None, 
                 neglect_only_autodep = False,
                 clean_cls = True):
        """
        Class constructor.

        Args:
            data (Data): data to analyse.
            min_lag (int): minimum time lag.
            max_lag (int): maximum time lag.
            val_condtest (CondIndTest): validation method.
            verbosity (CPLevel): verbosity level.
            alpha (float, optional): PCMCI significance level. Defaults to 0.05.
            resfolder (string, optional): result folder to create. Defaults to None.
            neglect_only_autodep (bool, optional): Bit for neglecting variables with only autodependency. Defaults to False.
            clean_cls (bool): Clean console bit. Default to True.
        """
        super().__init__(data, min_lag, max_lag, verbosity, alpha, resfolder, neglect_only_autodep, clean_cls)
        
        # build tigramite dataset
        vector = np.vectorize(float)
        d = vector(data.d)
        
        # init pcmci
        self.lpcmci = lpcmci(dataframe = pp.DataFrame(data = d, var_names = data.features),
                             cond_ind_test = val_condtest,
                             verbosity = verbosity.value)
        

    def run(self, link_assumptions = None) -> DAG:
        """
        Run causal discovery algorithm.

        Args:
            link_assumptions (dict, optional): prior knowledge on causal model links. Defaults to None.

        Returns:
            (DAG): estimated causal model.
        """
        CP.info('\n')
        CP.info(DASH)
        CP.info("Running Causal Discovery Algorithm")
        self.result = self.lpcmci.run_lpcmci(link_assumptions = link_assumptions,
                                             tau_max = self.max_lag,
                                             tau_min = self.min_lag,
                                             pc_alpha = self.alpha)
        
        self.CM = self._to_DAG()
            
        if self.resfolder is not None: self.logger.close()
        return self.CM
    
    
    def _to_DAG(self):
        """
        Re-elaborate the PCMCI result in a new dictionary.

        Returns:
            (DAG): lpcmci result re-elaborated.
        """
        vars = self.data.features
        tmp_dag = DAG(vars, self.min_lag, self.max_lag)
        tmp_dag.sys_context = dict()
        N, lags = self.result['graph'][0].shape
        for s in range(len(self.result['graph'])):
            for t in range(N):
                for lag in range(lags):
                    if self.result['graph'][s][t,lag] != '':
                        arrowtype = self.result['graph'][s][t,lag]
                        
                        if arrowtype == LinkType.Bidirected.value:
                            if ((vars[s], abs(lag)) in tmp_dag.g[vars[t]].sources and 
                                tmp_dag.g[t].sources[(vars[s], abs(lag))][TYPE] == LinkType.Bidirected.value):
                                continue
                            else:
                                tmp_dag.add_source(vars[t], 
                                                vars[s],
                                                self.result['val_matrix'][s][t,lag],
                                                self.result['p_matrix'][s][t,lag],
                                                lag,
                                                arrowtype)
                                
                        
                        elif arrowtype == LinkType.Uncertain.value:
                            if ((vars[t], abs(lag)) in tmp_dag.g[vars[s]].sources and 
                                tmp_dag.g[vars[s]].sources[(vars[t], abs(lag))][TYPE] == LinkType.Uncertain.value):
                                continue
                            else:
                                tmp_dag.add_source(vars[t], 
                                                vars[s],
                                                self.result['val_matrix'][s][t,lag],
                                                self.result['p_matrix'][s][t,lag],
                                                lag,
                                                arrowtype)
                                
                        
                        elif (arrowtype == LinkType.Directed.value or
                              arrowtype == LinkType.HalfUncertain.value):
                            tmp_dag.add_source(vars[t], 
                                            vars[s],
                                            self.result['val_matrix'][s][t,lag],
                                            self.result['p_matrix'][s][t,lag],
                                            lag,
                                            arrowtype)
        return tmp_dag


"""
This module provides the Static class.

Classes:
    Static: Subsamples data by taking one sample each step-samples.
"""

from causalflow.preprocessing.subsampling_methods.SubsamplingMethod import SubsamplingMethod, SSMode

class Static(SubsamplingMethod):
    """Subsample data by taking one sample each step-samples."""
    
    def __init__(self, step):
        """
        Class constructor.
        
        Args:
            step (int): integer subsampling step.

        Raises:
            ValueError: if step == None.
        """
        super().__init__(SSMode.Static)
        if step is None:
            raise ValueError("step not specified")
        self.step = step

    def run(self):
        """Run subsampler."""
        return range(0, len(self.df.values), self.step)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Implements the BaseGraph and Graph classes.
"""

import warnings
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from ._utils import (
    _get_unique_nodes,
    _get_angle,
    _get_point_along_spline,
    _get_tangent_at_point,
    _make_pretty,
    _rank,
    _normalize_numeric_argument,
    _normalize_color_argument,
    _normalize_shape_argument,
    _rescale_dict_values,
    _check_completeness,
    _get_optimal_offsets,
    _resample_spline,
    _get_total_pixels,
)
from ._node_layout import (
    get_fruchterman_reingold_layout,
    get_random_layout,
    get_sugiyama_layout,
    get_radial_tree_layout,
    get_circular_layout,
    get_linear_layout,
    get_bipartite_layout,
    get_multipartite_layout,
    get_shell_layout,
    get_community_layout,
    get_geometric_layout,
    _remove_node_overlap,
)
from ._edge_layout import (
    StraightEdgeLayout,
    CurvedEdgeLayout,
    BundledEdgeLayout,
    ArcDiagramEdgeLayout,
)

from ._artists import (
    Path,
    NodeArtist,
    CircularNodeArtist,
    RegularPolygonNodeArtist,
    EdgeArtist,
)
from ._parser import parse_graph, _parse_edge_list


BASE_SCALE = 1e-2
DEFAULT_COLOR = '#2c404c' # '#677e8c' # '#121f26' # '#23343f' # 'k',
RELATIVE_ARROW_HEAD_LENGTH = 3 # multiplier applied to edge width
RELATIVE_ARROW_HEAD_WIDTH = 2 # ditto


class BaseGraph(object):
    """The Graph base class.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    nodes : list or None, default None
        List of nodes. Required argument if any node in the graph is unconnected.
        If None, `nodes` is initialised to the set of the flattened `edges`.
    node_layout : str or dict, default 'spring'
        If a string, the node positions are computed using the indicated method:

        - 'random'       : place nodes in random positions;
        - 'circular'     : place nodes regularly spaced on a circle;
        - 'spring'       : place nodes using a force-directed layout (Fruchterman-Reingold algorithm);
        - 'dot'          : place nodes using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'radial'       : place nodes radially using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'community'    : place nodes such that nodes belonging to the same community are grouped together;
        - 'bipartite'    : place nodes regularly spaced on two parallel lines;
        - 'multipartite' : place nodes regularly spaced on several parallel lines;
        - 'shell'        : place nodes regularly spaced on concentric circles;
        - 'geometric'    : place nodes according to the length of the edges between them.

        If a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_random_layout`
        - :py:func:`get_circular_layout`
        - :py:func:`get_fruchterman_reingold_layout`
        - :py:func:`get_sugiyama_layout`
        - :py:func:`get_radial_tree_layout`
        - :py:func:`get_community_layout`
        - :py:func:`get_bipartite_layout`
        - :py:func:`get_multipartite_layout`
        - :py:func:`get_shell_layout`
        - :py:func:`get_geometric_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Edge color.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout : str or dict (default 'straight')
        If edge_layout is a string, determine the layout internally:

        - 'straight' : draw edges as straight lines
        - 'curved'   : draw edges as curved splines; the spline control points are optimised to avoid other nodes and edges
        - 'bundled'  : draw edges as edge bundles
        - 'arc'      : draw edges as arcs with a fixed curvature

        If edge_layout is a dict, the keys are edges and the values are edge paths
        in the form iterables of (x, y) tuples, the edge segments.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to edge layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_straight_edge_paths`
        - :py:func:`get_curved_edge_paths`
        - :py:func:`get_bundled_edge_paths`
        - :py:func:`get_arced_edge_paths`

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`Graph`, :py:class:`BaseMultiGraph`, :py:class:`BaseArcDiagram`

    """

    def __init__(self, edges,
                 nodes=None,
                 node_layout='spring',
                 node_layout_kwargs=None,
                 node_shape='o',
                 node_size=3.,
                 node_edge_width=0.5,
                 node_color='w',
                 node_edge_color=DEFAULT_COLOR,
                 node_alpha=1.0,
                 node_zorder=2,
                 node_labels=False,
                 node_label_offset=(0., 0.),
                 node_label_fontdict=None,
                 edge_width=1.,
                 edge_color=DEFAULT_COLOR,
                 tail_color=DEFAULT_COLOR,
                 edge_alpha=0.5,
                 edge_zorder=1,
                 arrows=False,
                 edge_layout='straight',
                 edge_layout_kwargs=None,
                 edge_labels=False,
                 edge_label_position=0.5,
                 edge_label_rotate=True,
                 edge_label_fontdict=None,
                 origin=(0., 0.),
                 scale=(1., 1.),
                 prettify=True,
                 ax=None,
                 *args, **kwargs
    ):
        self.ax = self._initialize_axis(ax)
        self.fig = self.ax.get_figure()

        self.edges = self._initialize_edges(edges)
        self.nodes = self._initialize_nodes(nodes)

        self._raise_warning_if_graph_too_large()

        # Convert all node and edge parameters to dictionaries.
        node_shape      = _normalize_shape_argument(node_shape, self.nodes, 'node_shape')
        node_size       = _normalize_numeric_argument(node_size, self.nodes, 'node_size')
        node_edge_width = _normalize_numeric_argument(node_edge_width, self.nodes, 'node_edge_width')
        node_color      = _normalize_color_argument(node_color, self.nodes, 'node_color')
        node_edge_color = _normalize_color_argument(node_edge_color, self.nodes, 'node_edge_color')
        node_alpha      = _normalize_numeric_argument(node_alpha, self.nodes, 'node_alpha')
        node_zorder     = _normalize_numeric_argument(node_zorder, self.nodes, 'node_zorder')
        edge_width      = _normalize_numeric_argument(edge_width, self.edges, 'edge_width')
        edge_color      = _normalize_color_argument(edge_color, self.edges, 'edge_color')
        edge_alpha      = _normalize_numeric_argument(edge_alpha, self.edges, 'edge_alpha')
        edge_zorder     = _normalize_numeric_argument(edge_zorder, self.edges, 'edge_zorder')

        # Rescale.
        node_size       = _rescale_dict_values(node_size, BASE_SCALE)
        node_edge_width = _rescale_dict_values(node_edge_width, BASE_SCALE)
        edge_width      = _rescale_dict_values(edge_width, BASE_SCALE)

        self.node_size = node_size
        self.edge_width = edge_width
        self.origin = origin
        self.scale = scale

        # Initialise node and edge layouts and draw elements.
        self.node_positions = self._initialize_node_layout(
            node_layout, node_layout_kwargs)
        self.node_artists = dict()
        self.draw_nodes(self.nodes, self.node_positions,
                        node_shape, self.node_size, node_edge_width,
                        node_color, node_edge_color, node_alpha, node_zorder)

        self.edge_layout = self._initialize_edge_layout(edge_layout, edge_layout_kwargs)
        self.edge_artists = dict()
        self.draw_edges(self.edges, self.edge_layout.edge_paths, edge_width,
            edge_color, edge_alpha, edge_zorder, arrows, self.node_artists, tail_color)

        # This function needs to be called before any font sizes are adjusted,
        # as the axis dimensions affect the effective font size.
        self.ax.autoscale_view()
        if prettify:
            _make_pretty(self.ax)

        if node_labels:
            if isinstance(node_labels, bool):
                node_labels = dict(zip(self.nodes, self.nodes))
            self.node_label_fontdict = self._initialize_node_label_fontdict(node_label_fontdict)
            self.autoscale_node_labels = self._set_autoscale_node_labels_flag(self.node_label_fontdict, node_label_offset)
            self.node_label_offset, self._recompute_node_label_offsets =\
                self._initialize_node_label_offset(node_labels, node_label_offset)
            if self._recompute_node_label_offsets:
                self._update_node_label_offsets()
            self.node_label_artists = dict()
            self.draw_node_labels(node_labels, self.node_label_fontdict)
            if self.autoscale_node_labels:
                self._rescale_node_labels()
                self.fig.canvas.mpl_connect('resize_event', self._on_resize)
        else:
            self.autoscale_node_labels = False

        if edge_labels:
            if isinstance(edge_labels, bool):
                edge_labels = dict(zip(self.edges, self.edges))
            self.edge_label_fontdict = self._initialize_edge_label_fontdict(edge_label_fontdict)
            self.edge_label_position = edge_label_position
            self.edge_label_rotate = edge_label_rotate
            self.edge_label_artists = dict()
            self.draw_edge_labels(edge_labels, self.edge_label_position,
                                  self.edge_label_rotate, self.edge_label_fontdict)


    def _initialize_edges(self, edges):
        return _parse_edge_list(edges)


    def _initialize_nodes(self, nodes):
        nodes_in_edges = _get_unique_nodes(self.edges)
        if nodes is None:
            return nodes_in_edges
        else:
            if set(nodes).issuperset(nodes_in_edges):
                return nodes
            else:
                msg = "There are some node IDs in the edgelist not present in `nodes`. "
                msg += "`nodes` has to be the superset of `edges`."
                msg += "\nThe following nodes are missing:"
                missing = set(nodes_in_edges) - set(nodes)
                for node in missing:
                    msg += f"\n\t{node}"
                raise ValueError(msg)


    def _raise_warning_if_graph_too_large(self):
        total_pixels_per_edge = _get_total_pixels(self.fig) / len(self.edges)
        if total_pixels_per_edge < 400:
            msg = "The graph may be too large to visualize meaningfully."
            msg += f" There are only available {int(total_pixels_per_edge)} pixels per edge."
            msg += " This number is unlikely to be sufficient to render the edges without many overlaps, resulting in a 'hairball'."
            msg += " For comparison, each letter in this sentence is likely rendered using 200-400 pixels."
            import warnings
            warnings.warn(msg)


    def _initialize_node_layout(self, node_layout, node_layout_kwargs):
        if node_layout_kwargs is None:
            node_layout_kwargs = dict()

        if isinstance(node_layout, str):
            if (node_layout == 'spring') or (node_layout == 'dot') or (node_layout == 'radial'):
                node_layout_kwargs.setdefault('node_size', self.node_size)
            return self._get_node_positions(
                node_layout, node_layout_kwargs, self.edges,
            )

        elif isinstance(node_layout, dict):
            _check_completeness(set(node_layout), set(self.nodes), 'node_layout')
            # TODO check that nodes are within bounding box set by origin and scale
            return node_layout


    def _get_node_positions(self, node_layout, node_layout_kwargs, edges):
        if len(self.nodes) == 1:
            return {self.nodes[0]: np.array([self.origin[0] + 0.5 * self.scale[0], self.origin[1] + 0.5 * self.scale[1]])}
        if node_layout == 'spring':
            node_positions = get_fruchterman_reingold_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
            if len(node_positions) > 3: # Qhull fails for 2 or less nodes
                node_positions = _remove_node_overlap(node_positions, node_size=self.node_size, origin=self.origin, scale=self.scale)
            return node_positions
        elif node_layout == 'community':
            node_positions = get_community_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
            if len(node_positions) > 3: # Qhull fails for 2 or less nodes
                node_positions = _remove_node_overlap(node_positions, node_size=self.node_size, origin=self.origin, scale=self.scale)
            return node_positions
        elif node_layout == 'circular':
            return get_circular_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        elif node_layout == 'linear':
            return get_linear_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        elif node_layout == 'bipartite':
            return get_bipartite_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        elif node_layout == 'multipartite':
            return get_multipartite_layout(
                edges, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        elif node_layout == 'shell':
            return get_shell_layout(
                edges, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        elif node_layout == 'dot':
            return get_sugiyama_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        elif node_layout == 'radial':
            return get_radial_tree_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        elif node_layout == 'random':
            return get_random_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        elif node_layout == 'geometric':
            return get_geometric_layout(
                edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        else:
            implemented = ['spring', 'community', 'circular',
                           'linear', 'bipartite', 'multipartite',
                           'shell', 'dot', 'radial', 'random',
                           'geometric']
            msg = f"Node layout {node_layout} not implemented. Available layouts are:"
            for method in implemented:
                msg += f"\n\t{method}"
            raise NotImplementedError(msg)


    def _initialize_edge_layout(self, edge_layout, edge_layout_kwargs):

        if edge_layout_kwargs is None:
            edge_layout_kwargs = dict()

        selfloops = [(source, target) for (source, target) in self.edges if source==target]
        if selfloops:
            if 'selfloop_radius' in edge_layout_kwargs:
                edge_layout_kwargs['selfloop_radius'] = \
                    _normalize_numeric_argument(edge_layout_kwargs['selfloop_radius'], selfloops, 'selfloop_radius')
            else:
                edge_layout_kwargs['selfloop_radius'] = \
                    {(node, node) : 1.5 * self.node_size[node] for node, _ in selfloops}

        if isinstance(edge_layout, str):
            if edge_layout == "straight":
                edge_layout = StraightEdgeLayout(self.edges, self.node_positions, **edge_layout_kwargs)
            elif edge_layout == "curved":
                edge_layout_kwargs.setdefault('node_size', self.node_size)
                edge_layout_kwargs.setdefault('origin', self.origin)
                edge_layout_kwargs.setdefault('scale', self.scale)
                edge_layout = CurvedEdgeLayout(self.edges, self.node_positions, **edge_layout_kwargs)
            elif edge_layout == "bundled":
                edge_layout = BundledEdgeLayout(self.edges, self.node_positions, **edge_layout_kwargs)
            elif edge_layout == "arc":
                edge_layout = ArcDiagramEdgeLayout(self.edges, self.node_positions, **edge_layout_kwargs)
            else:
                raise NotImplementedError(f"Variable edge_layout one of 'straight', 'curved', 'bundled', or 'arc', not {edge_layout}")
            edge_layout.get()

        elif isinstance(edge_layout, dict):
            _check_completeness(edge_layout, self.edges, 'edge_layout')
            # TODO check that edge paths are within bounding box given by origin and scale
            edge_paths = edge_layout
            edge_layout = StraightEdgeLayout(self.edges, self.node_positions, **edge_layout_kwargs)
            edge_layout.edge_paths.update(edge_paths)

        else:
            raise TypeError("Variable `edge_layout` either a string or a dict mapping edges to edge paths.")

        return edge_layout


    def _initialize_axis(self, ax):
        if ax is None:
            return plt.gca()
        elif isinstance(ax, mpl.axes.Axes):
            return ax
        else:
            raise TypeError(f"Variable 'ax' either None or a matplotlib axis instance. However, type(ax) is {type(ax)}.")


    def draw_nodes(self, nodes, node_positions, node_shape, node_size,
                   node_edge_width, node_color, node_edge_color, node_alpha,
                   node_zorder):
        """Draw or update node artists.

        Parameters
        ----------
        nodes : list
            List of nodes IDs.
        node_positions : dict
            Mapping of nodes to (x, y) positions.
        node_shape : dict
            Mapping of nodes to shapes.
            Specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
        node_size : dict
            Mapping of nodes to sizes.
        node_edge_width : dict
            Mapping of nodes to marker edge widths.
        node_color : dict
            Mapping of nodes to valid matplotlib color specifications.
        node_edge_color : dict
            Mapping of nodes to valid matplotlib color specifications.
        node_alpha : dict
            Mapping of nodes to node transparencies.
        node_zorder : dict
            Mapping of nodes to z-orders.

        Returns
        -------
        node_artists: dict
            Updates mapping of nodes to corresponding node artists.

        """

        for node in nodes:

            kwargs = dict(
                xy        = node_positions[node],
                size      = node_size[node],
                facecolor = node_color[node],
                edgecolor = node_edge_color[node],
                linewidth = node_edge_width[node],
                alpha     = node_alpha[node],
                zorder    = node_zorder[node],
            )

            shape = node_shape[node]

            if isinstance(shape, Path):
                node_artist = NodeArtist(shape, **kwargs)

            elif isinstance(shape, str):
                if shape == 'o':
                    node_artist = CircularNodeArtist(**kwargs)
                elif shape in 's^>v<dph8':
                    symbol_to_parameters = {
                        '^' : (3, 0),
                        '<' : (3, np.pi * 0.5),
                        'v' : (3, np.pi),
                        '>' : (3, np.pi * 1.5),
                        's' : (4, np.pi * 0.25),
                        'd' : (4, np.pi * 0.5),
                        'p' : (5, 0),
                        'h' : (6, 0),
                        '8' : (8, 0),
                    }
                    node_artist = RegularPolygonNodeArtist(*symbol_to_parameters[shape], **kwargs)
                else:
                    raise ValueError("Node shape should be one of: 'so^>v<dph8'. Current shape:{}".format(shape))
            else:
                raise ValueError("Node shape should be a matplotlob.Path instance or one of: 'so^>v<dph8'. Current shape:{}".format(shape))

            self.ax.add_patch(node_artist)

            if node in self.node_artists:
                self.node_artists[node].remove()
            self.node_artists[node] = node_artist


    def _update_node_artists(self, nodes):
        for node in nodes:
            self.node_artists[node].xy = self.node_positions[node]


    def draw_edges(self, edges, edge_path, edge_width, edge_color, edge_alpha,
                   edge_zorder, arrows, node_artists, tail_color):
        """Draw or update edge artists.

        Parameters
        ----------
        edges : list
            The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
        edge_path : dict
            Mapping of edges to arrays of (x, y) tuples, the edge path coordinates.
        edge_width : dict
            Mapping of edges to floats, the edge widths.
        edge_color : dict
            Mapping of edges to valid matplotlib color specifications, the edge colors.
        edge_alpha : dict
            Mapping of edges to floats, the edge transparencies.
        edge_zorder : dict
            Mapping of edges to ints, the edge z-order values.
        arrows : bool
            If True, draw edges with arrow heads.
        node_artists : dict
            Mapping of nodes to node artists. Required to offset edges from nodes.

        Returns
        -------
        self.edge_artists: dict
            Updates mapping of edges to corresponding edge artists.

        """

        for edge in edge_path:

            curved = False if (len(edge_path[edge]) == 2) else True

            source, target = edge
            if ((target, source) in edge_path) and (source != target): # i.e. bidirectional edges excluding self-loops
                if np.allclose(edge_path[(source, target)], edge_path[(target, source)][::-1]): # i.e. same path
                    shape = 'right' # i.e. plot half arrow / thin line shifted to the right
                else:
                    shape = 'full'
            else:
                shape = 'full'
            if arrows[edge]['h'] == '>':
                head_length = RELATIVE_ARROW_HEAD_LENGTH * edge_width[edge]
                head_width = RELATIVE_ARROW_HEAD_WIDTH * edge_width[edge]
            else:
                head_length = 0
                head_width = 0

            edge_artist = EdgeArtist(
                midline     = edge_path[edge],
                width       = edge_width[edge],
                facecolor   = edge_color[edge],
                alpha       = edge_alpha[edge],
                head_length = head_length,
                head_width  = head_width,
                edgecolor   = 'none',
                linewidth   = 0.,
                head_offset = node_artists[target].get_head_offset(edge_path[edge]),
                tail_offset = node_artists[source].get_tail_offset(edge_path[edge]),
                shape       = shape,
                curved      = curved,
                zorder      = edge_zorder[edge],
            )
            self.ax.add_patch(edge_artist)
            if arrows[edge]['h'] != '' and arrows[edge]['h'] != '>':
                edge_artist.draw_marker(self.ax, 'h', arrows[edge]['t'], edge_width[edge]/2, tail_color, tail_color, node_artists[source].size)
            if arrows[edge]['t'] != '':
                edge_artist.draw_marker(self.ax, 't', arrows[edge]['t'], edge_width[edge]/2, tail_color, tail_color, node_artists[source].size)

            if edge in self.edge_artists:
                self.edge_artists[edge].remove()
            self.edge_artists[edge] = edge_artist
            

    def _update_edge_artists(self, edge_paths=None):
        if edge_paths is None:
            edge_paths = self.edge_layout.edge_paths

        for edge, path in edge_paths.items():
            self.edge_artists[edge].update_midline(path)
            self.ax.draw_artist(self.edge_artists[edge])


    def _initialize_node_label_offset(self, node_labels, node_label_offset):
        if isinstance(node_label_offset, (int, float)):
            node_label_offset = {node : node_label_offset * self._get_vector_pointing_outwards(self.node_positions[node]) for node in node_labels}
            recompute = True
            return node_label_offset, recompute
        elif isinstance(node_label_offset, (tuple, list, np.ndarray)):
            if len(node_label_offset) == 2:
                node_label_offset = {node : node_label_offset for node in node_labels}
                recompute = False
                return node_label_offset, recompute
            else:
                msg = "If the variable `node_label_offset` is an iterable, it should have length 2."
                msg+= f"Current length: {len(node_label_offset)}."
                raise ValueError(msg)
        else:
            msg = "The variable `node_label_offset` has to be either a float, an int, a tuple, a list, or a numpy ndarray."
            msg += f"\nCurrent type: {type(node_label_offset)}."
            raise TypeError(msg)


    def _get_centroid(self):
        return np.mean([position for position in self.node_positions.values()], axis=0)


    def _get_vector_pointing_outwards(self, xy):
        centroid = self._get_centroid()
        delta = xy - centroid
        distance = np.linalg.norm(delta)
        unit_vector = delta / distance
        return unit_vector


    def _initialize_node_label_fontdict(self, node_label_fontdict):
        if node_label_fontdict is None:
            node_label_fontdict = dict()

        node_label_fontdict.setdefault('clip_on', False)
        node_label_fontdict.setdefault('zorder', np.inf)

        if 'ha' not in node_label_fontdict:
            node_label_fontdict.setdefault('horizontalalignment', 'center')

        if 'va' not in node_label_fontdict:
            node_label_fontdict.setdefault('verticalalignment', 'center')

        return node_label_fontdict


    def _set_autoscale_node_labels_flag(self, node_label_fontdict, node_label_offset):
        flag = False
        if ('size' not in node_label_fontdict) and ('fontsize' not in node_label_fontdict):
            if (node_label_fontdict.get('verticalalignment', 'center') == 'center') and \
               (node_label_fontdict.get('va', 'center') == 'center') and \
               (node_label_fontdict.get('horizonaltalignment', 'center') == 'center') and \
               (node_label_fontdict.get('ha', 'center') == 'center'):
                if np.all(np.isclose(node_label_offset, (0, 0))): # Labels are centered on node artists.
                    flag = True
        return flag


    def draw_node_labels(self, node_labels, node_label_fontdict):
        """Draw or update node labels.

        Parameters
        ----------
        node_labels : dict
           Mapping of nodes to strings, the node labels.
           Only nodes in the dictionary are labelled.
        node_label_offset: tuple, default (0., 0.)
            The (x, y) offset from node centre of label position.
        node_label_fontdict : dict
            Keyword arguments passed to matplotlib.text.Text.
            For a full list of available arguments see the matplotlib documentation.
            The following default values differ from the defaults for matplotlib.text.Text:

                - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
                - :code:`horizontalalignment` (default here: :code:`'center'`)
                - :code:`verticalalignment` (default here: :code:`'center'`)
                - :code:`clip_on` (default here: :code:`False`)

        Returns
        -------
        self.node_label_artists: dict
            Updates mapping of nodes to text objects, the node label artists.

        """

        for node, label in node_labels.items():
            x, y = self.node_positions[node]
            dx, dy = self.node_label_offset[node]
            artist = self.ax.text(x+dx, y+dy, label, **node_label_fontdict)

            if node in self.node_label_artists:
                self.node_label_artists[node].remove()
            self.node_label_artists[node] = artist


    def _rescale_node_labels(self, fudge_factor=0.75):
        maximum_font_sizes = [self.node_artists[node].get_maximum_fontsize(text_object) \
                              for node, text_object in self.node_label_artists.items()]
        font_size = np.min(maximum_font_sizes)
        for text_object in self.node_label_artists.values():
            text_object.set_size(fudge_factor * font_size)


    def _maximize_node_labels(self, fudge_factor=0.75):
        for node, text_object in self.node_label_artists.items():
            font_size = self.node_artists[node].get_maximum_fontsize(text_object)
            text_object.set_size(fudge_factor * font_size)


    def _on_resize(self, event):
        self._rescale_node_labels()


    def _update_node_label_positions(self):
        if self._recompute_node_label_offsets:
            self._update_node_label_offsets()

        for node, (dx, dy) in self.node_label_offset.items():
            x, y = self.node_positions[node]
            self.node_label_artists[node].set_position((x + dx, y + dy))


    def _update_node_label_offsets(self, total_samples_per_edge=100):
        anchors = np.array([self.node_positions[node] for node in self.node_label_offset.keys()])
        offsets = np.array(list(self.node_label_offset.values()))
        avoid = np.concatenate([_resample_spline(path, total_samples_per_edge) \
                                for path in self.edge_layout.edge_paths.values()], axis=0)
        optimal_offsets = _get_optimal_offsets(anchors, offsets, avoid)

        for ii, node in enumerate(self.node_label_offset):
            self.node_label_offset[node] = optimal_offsets[ii]


    def _initialize_edge_label_fontdict(self, edge_label_fontdict):
        if edge_label_fontdict is None:
            edge_label_fontdict = dict()

        edge_label_fontdict.setdefault('bbox', dict(color="white", pad=0))
        edge_label_fontdict.setdefault('horizontalalignment', 'center')
        edge_label_fontdict.setdefault('verticalalignment', 'center')
        edge_label_fontdict.setdefault('clip_on', False)
        edge_label_fontdict.setdefault('zorder', np.inf)
        return edge_label_fontdict


    def draw_edge_labels(self, edge_labels, edge_label_position,
                         edge_label_rotate, edge_label_fontdict):
        """Draw or update edge labels.

        Parameters
        ----------
        edge_labels : dict
            Mapping of edges to strings, the edge labels.
            Only edges in the dictionary are labelled.
        edge_label_position : float
            Relative position along the edge where the label is placed.

                - head   : 0.
                - centre : 0.5
                - tail   : 1.

        edge_label_rotate : bool
            If True, edge labels are rotated such that they have the same orientation as their corresponding edge.
            If False, edge labels are not rotated; the angle of the text is parallel to the axis.
        edge_label_fontdict : dict
            Keyword arguments passed to matplotlib.text.Text.

        Returns
        -------
        self.edge_label_artists: dict
            Updates mapping of edges to text objects, the edge label artists.

        """

        for edge, label in edge_labels.items():

            edge_artist = self.edge_artists[edge]

            x, y = _get_point_along_spline(edge_artist.midline, edge_label_position)

            if edge_label_rotate:

                # get tangent in degrees
                dx, dy = _get_tangent_at_point(edge_artist.midline, edge_label_position)
                angle = _get_angle(dx, dy, radians=False)

                # make label orientation "right-side-up"
                if angle > 90:
                    angle -= 180
                if angle < - 90:
                    angle += 180

            else:
                angle = None

            edge_label_artist = self.ax.text(x, y, label,
                                             rotation=angle,
                                             **edge_label_fontdict)

            if edge in self.edge_label_artists:
                self.edge_label_artists[edge].remove()
            self.edge_label_artists[edge] = edge_label_artist


    def _is_selfloop(self, edge):
        return True if edge[0] == edge[1] else False


    def _update_edge_label_positions(self, edges):
        for edge in edges:
            if edge in self.edge_label_artists:
                x, y = _get_point_along_spline(
                    self.edge_artists[edge].midline,
                    self.edge_label_position
                )
                self.edge_label_artists[edge].set_position((x, y))

                if self.edge_label_rotate:
                    dx, dy = _get_tangent_at_point(
                        self.edge_artists[edge].midline,
                        self.edge_label_position
                    )
                    angle = _get_angle(dx, dy, radians=False)
                    # make label orientation "right-side-up"
                    if angle > 90:
                        angle -= 180
                    if angle < -90:
                        angle += 180
                    # transform data coordinate angle to screen coordinate angle
                    trans_angle = self.ax.transData.transform_angles(np.array((angle,)), np.atleast_2d((x, y)))[0]
                    self.edge_label_artists[edge].set_rotation(trans_angle)


class Graph(BaseGraph):
    """Parses the given graph data object and initialises the :py:class:`BaseGraph` object.

    If the given graph includes edge weights, then these are mapped to colors using the `edge_cmap` parameter.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    node_layout : str or dict, default 'spring'
        If a string, the node positions are computed using the indicated method:

        - 'random'       : place nodes in random positions;
        - 'circular'     : place nodes regularly spaced on a circle;
        - 'spring'       : place nodes using a force-directed layout (Fruchterman-Reingold algorithm);
        - 'dot'          : place nodes using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'radial'       : place nodes radially using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'community'    : place nodes such that nodes belonging to the same community are grouped together;
        - 'bipartite'    : place nodes regularly spaced on two parallel lines;
        - 'multipartite' : place nodes regularly spaced on several parallel lines;
        - 'shell'        : place nodes regularly spaced on concentric circles;
        - 'geometric'    : place nodes according to the length of the edges between them.

        If a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_random_layout`
        - :py:func:`get_circular_layout`
        - :py:func:`get_fruchterman_reingold_layout`
        - :py:func:`get_sugiyama_layout`
        - :py:func:`get_radial_tree_layout`
        - :py:func:`get_community_layout`
        - :py:func:`get_bipartite_layout`
        - :py:func:`get_multipartite_layout`
        - :py:func:`get_shell_layout`
        - :py:func:`get_geometric_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_cmap : matplotlib color map (default 'RdGy')
        Color map used to map edge weights to edge colors. Should be diverging.
        If edge weights are strictly positive, weights are mapped to the
        left hand side of the color map with vmin=0 and vmax=np.max(weights).
        If edge weights are positive and negative, then weights are mapped
        to colors such that a weight of zero corresponds to the center of the
        color map; the boundaries are set to +/- the maximum absolute weight.
        If the graph is unweighted or the edge colors are specified explicitly,
        this parameter is ignored.
    edge_color : matplotlib color specification or dict, default DEFAULT_COLOR
        Edge color. If provided explicitly, overrides `edge_cmap`.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout : str or dict (default 'straight')
        If edge_layout is a string, determine the layout internally:

        - 'straight' : draw edges as straight lines
        - 'curved'   : draw edges as curved splines; the spline control points are optimised to avoid other nodes and edges
        - 'bundled'  : draw edges as edge bundles
        - 'arc'      : draw edges as arcs with a fixed curvature

        If edge_layout is a dict, the keys are edges and the values are edge paths
        in the form iterables of (x, y) tuples, the edge segments.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to edge layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_straight_edge_paths`
        - :py:func:`get_curved_edge_paths`
        - :py:func:`get_bundled_edge_paths`
        - :py:func:`get_arced_edge_paths`

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`BaseGraph`, :py:class:`InteractiveGraph`, :py:class:`MultiGraph`, :py:class:`ArcDiagram`

    """

    def __init__(self, graph, edge_cmap='RdGy', *args, **kwargs):

        # Accept a variety of formats for 'graph' and convert to common denominator.
        nodes, edges, edge_weight = self._parse_input(graph)
        kwargs.setdefault('nodes', nodes)

        # Color and reorder edges for weighted graphs.
        if edge_weight:
            # If the graph is weighted, we want to visualise the weights using color.
            # Edge width is another popular choice when visualising weighted networks,
            # but if the variance in weights is large, this typically results in less
            # visually pleasing results.
            edge_color = _get_color(edge_weight, cmap=edge_cmap)

            # Plotting darker edges over lighter edges typically results in visually
            # more pleasing results. Here we hence specify the relative order in
            # which edges are plotted according to the color of the edge.
            edge_zorder = _get_zorder(edge_color)
            node_zorder = np.max(list(edge_zorder.values())) + 1

            kwargs.setdefault('edge_color', edge_color)
            kwargs.setdefault('edge_zorder', edge_zorder)
            kwargs.setdefault('node_zorder', node_zorder)

        super().__init__(edges, *args, **kwargs)


    def _parse_input(self, graph):
        return parse_graph(graph)


def _get_color(mydict, cmap='RdGy', vmin=None, vmax=None):
    """Map positive and negative floats to a diverging colormap, such that

    - the midpoint of the colormap corresponds to a value of 0., and
    - values above and below the midpoint are mapped linearly and in equal measure
      to increases in color intensity.

    Parameters
    ----------
    mydict: dict
        Mapping of graph element (node, edge) to a float.
        For example (source, target) : edge weight.
    cmap: str, default 'RdGy'
        Matplotlib colormap specification.
    vmin, vmax: float or None, default None
        Minimum and maximum float corresponding to the dynamic range of the colormap.

    Returns
    -------
    newdict: dict
        Mapping of graph elements to RGBA tuples.

    """

    keys = mydict.keys()
    values = np.array(list(mydict.values()), dtype=float)

    # apply vmin, vmax
    if vmin or vmax:
        values = np.clip(values, vmin, vmax)

    def abs(value):
        try:
            return np.abs(value)
        except TypeError as e: # value is probably None
            if isinstance(value, type(None)):
                return 0
            else:
                raise e

    # rescale values such that
    #  - the colormap midpoint is at zero-value, and
    #  - negative and positive values have comparable intensity values
    values /= np.nanmax([np.nanmax(np.abs(values)), abs(vmax), abs(vmin)]) # [-1, 1]
    values += 1. # [0, 2]
    values /= 2. # [0, 1]

    # convert value to color
    mapper = mpl.cm.ScalarMappable(cmap=cmap)
    mapper.set_clim(vmin=0., vmax=1.)
    colors = mapper.to_rgba(values)

    return {key: color for (key, color) in zip(keys, colors)}


def _get_zorder(color_dict):
    """Reorder plot elements such that darker items are plotted last and hence most prominent in the graph.
    This assumes that the background is white.

    """
    intensities = [rgba_to_grayscale(*v) for v in color_dict.values()]
    zorder = _rank(intensities)
    zorder = np.max(zorder) - zorder # reverse order as greater values correspond to lighter colors
    return {key: index for key, index in zip(color_dict.keys(), zorder)}


def rgba_to_grayscale(r, g, b, a=1):
    """Convert RGBA values to grayscale.

    Notes
    -----
    Adapted from: https://stackoverflow.com/a/689547/2912349

    """

    return (0.299 * r + 0.587 * g + 0.114 * b) * a
